@@ -42,31 +42,32 @@ void pl_gpu_print_info(const struct pl_gpu *gpu, enum pl_log_level lev)
     PL_MSG(gpu, lev, "    Limits:");
 
 #define LOG(fmt, field) \
-    PL_MSG(gpu, lev, "      %-26s " fmt, #field ":", gpu->limits.field)
+    PL_MSG(gpu, lev, "      %-26s %" fmt, #field ":", gpu->limits.field)
 
-    LOG("%d", max_tex_1d_dim);
-    LOG("%d", max_tex_2d_dim);
-    LOG("%d", max_tex_3d_dim);
-    LOG("%zu", max_pushc_size);
-    LOG("%zu", max_xfer_size);
-    LOG("%zu", max_ubo_size);
-    LOG("%zu", max_ssbo_size);
-    LOG("%d", min_gather_offset);
-    LOG("%d", max_gather_offset);
+    LOG(PRIu32, max_tex_1d_dim);
+    LOG(PRIu32, max_tex_2d_dim);
+    LOG(PRIu32, max_tex_3d_dim);
+    LOG("zu", max_pushc_size);
+    LOG("zu", max_xfer_size);
+    LOG("zu", max_ubo_size);
+    LOG("zu", max_ssbo_size);
+    LOG(PRIu64, max_buffer_texels);
+    LOG(PRId16, min_gather_offset);
+    LOG(PRId16, max_gather_offset);
 
     if (gpu->caps & PL_GPU_CAP_COMPUTE) {
-        LOG("%zu", max_shmem_size);
-        LOG("%d", max_group_threads);
-        LOG("%d", max_group_size[0]);
-        LOG("%d", max_group_size[1]);
-        LOG("%d", max_group_size[2]);
-        LOG("%d", max_dispatch[0]);
-        LOG("%d", max_dispatch[1]);
-        LOG("%d", max_dispatch[2]);
+        LOG("zu", max_shmem_size);
+        LOG(PRIu32, max_group_threads);
+        LOG(PRIu32, max_group_size[0]);
+        LOG(PRIu32, max_group_size[1]);
+        LOG(PRIu32, max_group_size[2]);
+        LOG(PRIu32, max_dispatch[0]);
+        LOG(PRIu32, max_dispatch[1]);
+        LOG(PRIu32, max_dispatch[2]);
     }
 
-    LOG("%d", align_tex_xfer_stride);
-    LOG("%zu", align_tex_xfer_offset);
+    LOG(PRIu32, align_tex_xfer_stride);
+    LOG("zu", align_tex_xfer_offset);
 #undef LOG
 }
 
@@ -481,14 +482,12 @@ void pl_tex_blit(const struct pl_gpu *gpu,
 size_t pl_tex_transfer_size(const struct pl_tex_transfer_params *par)
 {
     const struct pl_tex *tex = par->tex;
+    int w = pl_rect_w(par->rc), h = pl_rect_h(par->rc), d = pl_rect_d(par->rc);
 
-    int texels;
-    switch (pl_tex_params_dimension(tex->params)) {
-    case 1: texels = pl_rect_w(par->rc); break;
-    case 2: texels = pl_rect_h(par->rc) * par->stride_w; break;
-    case 3: texels = pl_rect_d(par->rc) * par->stride_w * par->stride_h; break;
-    }
-
+    // This generates the absolute bare minimum size of a buffer required to
+    // hold the data of a texture upload/download, by including stride padding
+    // only where strictly necessary.
+    int texels = ((d - 1) * par->stride_h + (h - 1)) * par->stride_w + w;
     return texels * tex->params.format->texel_size;
 }
 
@@ -507,9 +506,9 @@ static void fix_tex_transfer(const struct pl_gpu *gpu,
         rc.z1 = tex->params.d;
 
     if (!params->stride_w)
-        params->stride_w = tex->params.w;
+        params->stride_w = pl_rect_w(rc);
     if (!params->stride_h)
-        params->stride_h = tex->params.h;
+        params->stride_h = pl_rect_h(rc);
 
     // Sanitize superfluous coordinates for the benefit of the GPU
     strip_coords(tex, &rc);
@@ -547,7 +546,9 @@ static void fix_tex_transfer(const struct pl_gpu *gpu,
     if (params->buf) {
         const struct pl_buf *buf = params->buf;
         size_t size = pl_tex_transfer_size(params);
+        size_t texel = tex->params.format->texel_size;
         pl_assert(params->buf_offset == PL_ALIGN2(params->buf_offset, 4));
+        pl_assert(params->buf_offset == PL_ALIGN(params->buf_offset, texel));
         pl_assert(params->buf_offset + size <= buf->params.size);
     }
 #endif
@@ -1099,7 +1100,7 @@ void pl_buf_pool_uninit(const struct pl_gpu *gpu, struct pl_buf_pool *pool)
         pl_buf_destroy(gpu, &pool->buffers[i]);
 
     talloc_free(pool->buffers);
-    *pool = (struct pl_buf_pool){0};
+    *pool = (struct pl_buf_pool) {0};
 }
 
 static bool pl_buf_params_compatible(const struct pl_buf_params *new,
