@@ -72,12 +72,15 @@ struct pl_renderer {
     bool disable_linear_sdr; // disable linear scaling for SDR signals
     bool disable_blending;   // disable blending for the target/fbofmt
     bool disable_overlay;    // disable rendering overlays
+    bool disable_3dlut;      // disable usage of a 3DLUT
 
     // Shader resource objects and intermediate textures (FBOs)
     struct pl_shader_obj *peak_detect_state;
     struct pl_shader_obj *dither_state;
+    struct pl_shader_obj *lut3d_state;
     const struct pl_tex *main_scale_fbo;
     const struct pl_tex *deband_fbos[PLANE_COUNT];
+    const struct pl_tex *output_fbo;
     struct sampler samplers[SCALER_COUNT];
     struct sampler *osd_samplers;
     int num_osd_samplers;
@@ -171,10 +174,12 @@ void pl_renderer_destroy(struct pl_renderer **p_rr)
     pl_tex_destroy(rr->gpu, &rr->main_scale_fbo);
     for (int i = 0; i < PL_ARRAY_SIZE(rr->deband_fbos); i++)
         pl_tex_destroy(rr->gpu, &rr->deband_fbos[i]);
+    pl_tex_destroy(rr->gpu, &rr->output_fbo);
 
     // Free all shader resource objects
     pl_shader_obj_destroy(&rr->peak_detect_state);
     pl_shader_obj_destroy(&rr->dither_state);
+    pl_shader_obj_destroy(&rr->lut3d_state);
 
     // Free all samplers
     for (int i = 0; i < PL_ARRAY_SIZE(rr->samplers); i++)
@@ -217,24 +222,32 @@ struct img {
     int comps;
 };
 
+// Returns the texture parameters an `img` would have if finalized
+static inline struct pl_tex_params img_params(struct pl_renderer *rr,
+                                              struct img *img,
+                                              const struct pl_fmt *fmt)
+{
+    return (struct pl_tex_params) {
+            .w = img->w,
+            .h = img->h,
+            .format = fmt,
+            .sampleable = true,
+            .renderable = true,
+            // Just enable what we can
+            .storable   = !!(fmt->caps & PL_FMT_CAP_STORABLE),
+            .sample_mode = (fmt->caps & PL_FMT_CAP_LINEAR)
+                                ? PL_TEX_SAMPLE_LINEAR
+                                : PL_TEX_SAMPLE_NEAREST,
+    };
+}
+
 static const struct pl_tex *finalize_img(struct pl_renderer *rr,
                                          struct img *img,
                                          const struct pl_fmt *fmt,
                                          const struct pl_tex **tex)
 {
-    bool ok = pl_tex_recreate(rr->gpu, tex, &(struct pl_tex_params) {
-        .w = img->w,
-        .h = img->h,
-        .format = fmt,
-        .sampleable = true,
-        .renderable = true,
-        // Just enable what we can
-        .storable   = !!(fmt->caps & PL_FMT_CAP_STORABLE),
-        .sample_mode = (fmt->caps & PL_FMT_CAP_LINEAR)
-                            ? PL_TEX_SAMPLE_LINEAR
-                            : PL_TEX_SAMPLE_NEAREST,
-    });
-
+    struct pl_tex_params tex_params = img_params(rr, img, fmt);
+    bool ok = pl_tex_recreate(rr->gpu, tex, &tex_params);
     if (!ok) {
         PL_ERR(rr, "Failed creating FBO texture! Disabling advanced rendering..");
         rr->fbofmt = NULL;
@@ -257,51 +270,104 @@ struct pass_state {
     struct img cur_img;
 };
 
+enum sampler_type {
+    SAMPLER_DIRECT, // texture()'s built in sampling
+    SAMPLER_BICUBIC, // fast bicubic scaling
+    SAMPLER_COMPLEX, // complex custom filters
+};
+
+enum sampler_dir {
+    SAMPLER_NOOP, // 1:1 scaling
+    SAMPLER_UP,   // upscaling
+    SAMPLER_DOWN, // downscaling
+};
+
+struct sampler_info {
+    const struct pl_filter_config *config; // if applicable
+    enum sampler_type type;
+    enum sampler_dir dir;
+};
+
+static struct sampler_info sample_src_info(struct pl_renderer *rr,
+                                           const struct pl_sample_src *src,
+                                           const struct pl_render_params *params)
+{
+    struct sampler_info info;
+
+    float rx = src->new_w / fabs(pl_rect_w(src->rect));
+    float ry = src->new_h / fabs(pl_rect_h(src->rect));
+    if (rx < 1.0 - 1e-6 || ry < 1.0 - 1e-6) {
+        info.dir = SAMPLER_DOWN;
+        info.config = params->downscaler;
+    } else if (rx > 1.0 + 1e-6 || ry > 1.0 + 1e-6) {
+        info.dir = SAMPLER_UP;
+        info.config = params->upscaler;
+    } else {
+        info.dir = SAMPLER_NOOP;
+        info.type = SAMPLER_DIRECT;
+        info.config = NULL;
+        return info;
+    }
+
+    bool is_linear = src->tex->params.sample_mode == PL_TEX_SAMPLE_LINEAR;
+    if (!rr->fbofmt || rr->disable_sampling || !info.config) {
+        info.type = is_linear ? SAMPLER_BICUBIC : SAMPLER_DIRECT;
+    } else {
+        info.type = SAMPLER_COMPLEX;
+
+        // Try using faster replacements for GPU built-in scalers
+        bool can_fast = info.config == params->upscaler ||
+                        params->skip_anti_aliasing;
+        if (can_fast && !params->disable_builtin_scalers) {
+            if (is_linear && info.config == &pl_filter_bicubic)
+                info.type = SAMPLER_BICUBIC;
+            if (is_linear && info.config == &pl_filter_triangle)
+                info.type = SAMPLER_DIRECT;
+            if (!is_linear && info.config == &pl_filter_box)
+                info.type = SAMPLER_DIRECT;
+        }
+    }
+
+    return info;
+}
+
 static void dispatch_sampler(struct pl_renderer *rr, struct pl_shader *sh,
-                             float rx, float ry, struct sampler *sampler,
+                             struct sampler *sampler,
                              const struct pl_render_params *params,
                              const struct pl_sample_src *src)
 {
-    bool is_linear = src->tex->params.sample_mode == PL_TEX_SAMPLE_LINEAR;
-    bool no_samplers = !params->upscaler && !params->downscaler;
-    if (!rr->fbofmt || rr->disable_sampling || !sampler || no_samplers)
+    if (!sampler)
         goto fallback;
 
-    const struct pl_filter_config *config = NULL;
-    struct pl_shader_obj **lut;
-    const struct pl_tex **sep_fbo;
-
-    rx = fabs(rx);
-    ry = fabs(ry);
-
-    if (rx < 1.0 - 1e-6 || ry < 1.0 - 1e-6) {
-        config = params->downscaler;
+    struct sampler_info info = sample_src_info(rr, src, params);
+    struct pl_shader_obj **lut = NULL;
+    const struct pl_tex **sep_fbo = NULL;
+    switch (info.dir) {
+    case SAMPLER_NOOP:
+        goto fallback;
+    case SAMPLER_DOWN:
         lut = &sampler->downscaler_state;
         sep_fbo = &sampler->sep_fbo_down;
-    } else if (rx > 1.0 + 1e-6 || ry > 1.0 + 1e-6) {
-        config = params->upscaler;
+        break;
+    case SAMPLER_UP:
         lut = &sampler->upscaler_state;
         sep_fbo = &sampler->sep_fbo_up;
-    } else { // no scaling
-        goto direct;
+        break;
     }
 
-    if (!config)
+    switch (info.type) {
+    case SAMPLER_DIRECT:
         goto fallback;
-
-    // Try using faster replacements for GPU built-in scalers
-    bool can_fast = config == params->upscaler || params->skip_anti_aliasing;
-    if (can_fast && !params->disable_builtin_scalers) {
-        if (is_linear && config == &pl_filter_bicubic)
-            goto fallback; // the bicubic check will succeed
-        if (is_linear && config == &pl_filter_triangle)
-            goto direct;
-        if (!is_linear && config == &pl_filter_box)
-            goto direct;
+    case SAMPLER_BICUBIC:
+        pl_shader_sample_bicubic(sh, src);
+        return;
+    case SAMPLER_COMPLEX:
+        break; // continue below
     }
 
+    pl_assert(lut && sep_fbo);
     struct pl_sample_filter_params fparams = {
-        .filter      = *config,
+        .filter      = *info.config,
         .lut_entries = params->lut_entries,
         .cutoff      = params->polar_cutoff,
         .antiring    = params->antiringing_strength,
@@ -311,7 +377,7 @@ static void dispatch_sampler(struct pl_renderer *rr, struct pl_shader *sh,
     };
 
     bool ok;
-    if (config->polar) {
+    if (info.config->polar) {
         ok = pl_shader_sample_polar(sh, src, &fparams);
     } else {
         struct pl_shader *tsh = pl_dispatch_begin(rr->dp);
@@ -329,7 +395,7 @@ static void dispatch_sampler(struct pl_renderer *rr, struct pl_shader *sh,
 
         struct pl_sample_src src2 = *src;
         src2.tex = finalize_img(rr, &img, rr->fbofmt, sep_fbo);
-        ok = pl_shader_sample_ortho(sh, PL_SEP_HORIZ, &src2, &fparams);
+        ok = src2.tex && pl_shader_sample_ortho(sh, PL_SEP_HORIZ, &src2, &fparams);
     }
 
 done:
@@ -342,13 +408,6 @@ done:
     return;
 
 fallback:
-    // Use bicubic sampling if supported
-    if (rr->fbofmt && is_linear) {
-        pl_shader_sample_bicubic(sh, src);
-        return;
-    }
-
-direct:
     // If all else fails, fall back to bilinear/nearest
     pl_shader_sample_direct(sh, src);
 }
@@ -401,15 +460,12 @@ static void draw_overlays(struct pl_renderer *rr, const struct pl_tex *fbo,
             },
         };
 
-        float rx = (float) src.new_w / src.tex->params.w,
-              ry = (float) src.new_h / src.tex->params.h;
-
         struct sampler *sampler = &rr->osd_samplers[n];
         if (params->disable_overlay_sampling)
             sampler = NULL;
 
         struct pl_shader *sh = pl_dispatch_begin(rr->dp);
-        dispatch_sampler(rr, sh, rx, ry, sampler, params, &src);
+        dispatch_sampler(rr, sh, sampler, params, &src);
 
         GLSL("vec4 osd_color;\n");
         for (int c = 0; c < src.components; c++) {
@@ -461,39 +517,83 @@ static void draw_overlays(struct pl_renderer *rr, const struct pl_tex *fbo,
     }
 }
 
-static void deband_plane(struct pl_renderer *rr, struct pl_plane *plane,
-                         const struct pl_tex **fbo,
-                         const struct pl_render_params *params)
-{
-    if (!rr->fbofmt || rr->disable_debanding || !params->deband_params)
-        return;
+// `deband_src` results
+enum {
+    DEBAND_NOOP = 0, // no debanding was performing
+    DEBAND_NORMAL,   // debanding was performed, the plane should still be scaled
+    DEBAND_SCALED,   // debanding took care of scaling as well
+};
 
-    const struct pl_tex *tex = plane->texture;
-    if (tex->params.sample_mode != PL_TEX_SAMPLE_LINEAR) {
+static int deband_src(struct pl_renderer *rr, struct pl_shader *psh,
+                      struct pl_sample_src *psrc, const struct pl_tex **fbo,
+                      const struct pl_render_params *params)
+{
+    if (rr->disable_debanding || !params->deband_params)
+        return DEBAND_NOOP;
+
+    if (psrc->tex->params.sample_mode != PL_TEX_SAMPLE_LINEAR) {
         PL_WARN(rr, "Debanding requires uploaded textures to be linearly "
                 "sampleable (params.sample_mode = PL_TEX_SAMPLE_LINEAR)! "
                 "Disabling debanding..");
         rr->disable_debanding = true;
-        return;
+        return DEBAND_NOOP;
     }
 
-    struct pl_shader *sh = pl_dispatch_begin(rr->dp);
-    pl_shader_deband(sh, tex, params->deband_params);
+    // The debanding shader can replace direct GPU sampling
+    bool deband_scales = sample_src_info(rr, psrc, params).type == SAMPLER_DIRECT;
+
+    struct pl_shader *sh = psh;
+    struct pl_sample_src *src = psrc;
+    if (!deband_scales) {
+        // Only sample/deband the relevant cut-out, but round it to the nearest
+        // integer to avoid doing fractional scaling
+        struct pl_sample_src fixed = *src;
+        fixed.rect.x0 = floorf(fixed.rect.x0);
+        fixed.rect.y0 = floorf(fixed.rect.y0);
+        fixed.rect.x1 = ceilf(fixed.rect.x1);
+        fixed.rect.y1 = ceilf(fixed.rect.y1);
+        fixed.new_w = pl_rect_w(fixed.rect);
+        fixed.new_h = pl_rect_h(fixed.rect);
+        src = &fixed;
+
+        if (fixed.new_w == psrc->new_w &&
+            fixed.new_h == psrc->new_h &&
+            pl_rect2d_eq(fixed.rect, psrc->rect))
+        {
+            // If there's nothing left to be done (i.e. we're already rendering
+            // an exact integer crop without scaling), also skip the scalers
+            deband_scales = true;
+        } else {
+            sh = pl_dispatch_begin(rr->dp);
+        }
+    }
+
+    pl_shader_deband(sh, src, params->deband_params);
+
+    if (deband_scales)
+        return DEBAND_SCALED;
 
     struct img img = {
         .sh = sh,
-        .w  = tex->params.w,
-        .h  = tex->params.h,
+        .w  = src->new_w,
+        .h  = src->new_h,
     };
 
     const struct pl_tex *new = finalize_img(rr, &img, rr->fbofmt, fbo);
     if (!new) {
         PL_ERR(rr, "Failed dispatching debanding shader.. disabling debanding!");
         rr->disable_debanding = true;
-        return;
+        return DEBAND_NOOP;
     }
 
-    plane->texture = new;
+    // Update the original pl_sample_src to point to the new texture
+    psrc->tex = new;
+    psrc->rect.x0 -= src->rect.x0;
+    psrc->rect.y0 -= src->rect.y0;
+    psrc->rect.x1 -= src->rect.x0;
+    psrc->rect.y1 -= src->rect.y0;
+    psrc->scale = 1.0;
+    return DEBAND_NORMAL;
 }
 
 // This scales and merges all of the source images, and initializes the cur_img.
@@ -560,10 +660,12 @@ static bool pass_read_image(struct pl_renderer *rr, struct pass_state *pass,
     pl_assert(target_w > 0 && target_h > 0);
 
     bool has_alpha = false;
+    struct pl_color_repr repr = image->repr;
+    float scale = pl_color_repr_normalize(&repr);
+
     for (int i = 0; i < image->num_planes; i++) {
         struct pl_shader *psh = pl_dispatch_begin(rr->dp);
         struct pl_plane *plane = &planes[i];
-        deband_plane(rr, plane, &rr->deband_fbos[i], params);
 
         // Compute the source shift/scale relative to the reference size
         float pw = plane->texture->params.w,
@@ -582,6 +684,7 @@ static bool pass_read_image(struct pl_renderer *rr, struct pass_state *pass,
         struct pl_sample_src src = {
             .tex        = plane->texture,
             .components = plane->components,
+            .scale      = scale,
             .new_w      = target_w,
             .new_h      = target_h,
             .rect       = {
@@ -592,10 +695,8 @@ static bool pass_read_image(struct pl_renderer *rr, struct pass_state *pass,
             },
         };
 
-        // FIXME: in theory, we could reuse the debanding result from
-        // `deband_plane` if available, instead of having to dispatch a no-op
-        // shader, for trivial sampling cases (no scaling or shifting)
-        dispatch_sampler(rr, psh, rrx, rry, &rr->samplers[i], params, &src);
+        if (deband_src(rr, psh, &src, &rr->deband_fbos[i], params) != DEBAND_SCALED)
+            dispatch_sampler(rr, psh, &rr->samplers[i], params, &src);
 
         ident_t sub = sh_subpass(sh, psh);
         if (!sub) {
@@ -615,7 +716,7 @@ static bool pass_read_image(struct pl_renderer *rr, struct pass_state *pass,
             if (plane->component_mapping[c] < 0)
                 continue;
             GLSL("color[%d] = tmp[%d];\n", plane->component_mapping[c],
-                 plane->texture->params.format->sample_order[c]);
+                 src.tex->params.format->sample_order[c]);
 
             has_alpha |= plane->component_mapping[c] == PLANE_A;
         }
@@ -631,7 +732,7 @@ static bool pass_read_image(struct pl_renderer *rr, struct pass_state *pass,
         .sh     = sh,
         .w      = target_w,
         .h      = target_h,
-        .repr   = image->repr,
+        .repr   = repr,
         .color  = image->color,
         .comps  = has_alpha ? 4 : 3,
         .rect   = {
@@ -653,46 +754,46 @@ static bool pass_scale_main(struct pl_renderer *rr, struct pass_state *pass,
                             const struct pl_render_target *target,
                             const struct pl_render_params *params)
 {
-    struct img *img = &pass->cur_img;
-    int target_w = abs(pl_rect_w(target->dst_rect)),
-        target_h = abs(pl_rect_h(target->dst_rect));
-
-    float src_w = pl_rect_w(image->src_rect),
-          src_h = pl_rect_h(image->src_rect);
-    pl_assert(src_w > 0 && src_h > 0);
-
-    float rx = target_w / src_w,
-          ry = target_h / src_h;
-
     if (!rr->fbofmt) {
         PL_TRACE(rr, "Skipping main scaler (no FBOs)");
         return true;
     }
 
-    if ((!params->upscaler && !params->downscaler) || rr->disable_sampling) {
-        PL_TRACE(rr, "Skipping main scaler (no samplers)");
+    struct img *img = &pass->cur_img;
+    struct pl_sample_src src = {
+        .tex        = &(struct pl_tex) { .params = img_params(rr, img, rr->fbofmt) },
+        .components = img->comps,
+        .new_w      = abs(pl_rect_w(target->dst_rect)),
+        .new_h      = abs(pl_rect_h(target->dst_rect)),
+        .rect       = img->rect,
+    };
+
+    struct sampler_info info = sample_src_info(rr, &src, params);
+    bool need_osd = image->num_overlays > 0;
+    if (info.type == SAMPLER_DIRECT && !need_osd) {
+        PL_TRACE(rr, "Skipping main scaler (free sampling)");
         return true;
     }
 
-    bool downscaling = rx < 1.0 - 1e-6 || ry < 1.0 - 1e-6;
-    bool upscaling = !downscaling && (rx > 1.0 + 1e-6 || ry > 1.0 + 1e-6);
-    bool need_osd = image->num_overlays > 0;
-
-    if (!downscaling && !upscaling && !need_osd) {
+    if (info.dir == SAMPLER_NOOP && !need_osd) {
         PL_TRACE(rr, "Skipping main scaler (would be no-op)");
         return true;
     }
 
-    bool use_sigmoid = upscaling && params->sigmoid_params;
-    bool use_linear  = use_sigmoid || downscaling;
+    bool use_sigmoid = info.dir == SAMPLER_UP && params->sigmoid_params;
+    bool use_linear  = use_sigmoid || info.dir == SAMPLER_DOWN;
 
-    // Hard-disable both sigmoidization and linearization when requested
-    if (params->disable_linear_scaling)
+    // Hard-disable both sigmoidization and linearization when required
+    if (params->disable_linear_scaling || rr->disable_linear_sdr)
         use_sigmoid = use_linear = false;
 
     // Avoid sigmoidization for HDR content because it clips to [0,1]
-    if (pl_color_transfer_is_hdr(img->color.transfer))
+    if (pl_color_transfer_is_hdr(img->color.transfer)) {
         use_sigmoid = false;
+        // Also disable linearization if necessary
+        if (rr->disable_linear_hdr)
+            use_linear = false;
+    }
 
     if (use_linear) {
         pl_shader_linearize(img->sh, img->color.transfer);
@@ -702,14 +803,7 @@ static bool pass_scale_main(struct pl_renderer *rr, struct pass_state *pass,
     if (use_sigmoid)
         pl_shader_sigmoidize(img->sh, params->sigmoid_params);
 
-    struct pl_sample_src src = {
-        .tex        = finalize_img(rr, img, rr->fbofmt, &rr->main_scale_fbo),
-        .components = img->comps,
-        .new_w      = target_w,
-        .new_h      = target_h,
-        .rect       = img->rect,
-    };
-
+    src.tex = finalize_img(rr, img, rr->fbofmt, &rr->main_scale_fbo);
     if (!src.tex)
         return false;
 
@@ -718,11 +812,11 @@ static bool pass_scale_main(struct pl_renderer *rr, struct pass_state *pass,
                   img->color, use_sigmoid, NULL, params);
 
     struct pl_shader *sh = pl_dispatch_begin(rr->dp);
-    dispatch_sampler(rr, sh, rx, ry, &rr->samplers[SCALER_MAIN], params, &src);
+    dispatch_sampler(rr, sh, &rr->samplers[SCALER_MAIN], params, &src);
     pass->cur_img = (struct img) {
         .sh     = sh,
-        .w      = target_w,
-        .h      = target_h,
+        .w      = src.new_w,
+        .h      = src.new_h,
         .repr   = img->repr,
         .color  = img->color,
         .comps  = img->comps,
@@ -735,15 +829,84 @@ static bool pass_scale_main(struct pl_renderer *rr, struct pass_state *pass,
 }
 
 static bool pass_output_target(struct pl_renderer *rr, struct pass_state *pass,
+                               const struct pl_image *image,
                                const struct pl_render_target *target,
                                const struct pl_render_params *params)
 {
     const struct pl_tex *fbo = target->fbo;
+    struct pl_shader *sh = pass->cur_img.sh;
 
     // Color management
-    struct pl_shader *sh = pass->cur_img.sh;
-    pl_shader_color_map(sh, params->color_map_params, pass->cur_img.color,
-                        target->color, &rr->peak_detect_state, false);
+    bool use_3dlut = image->profile.data || target->profile.data ||
+                     params->force_3dlut;
+    if (rr->disable_3dlut)
+        use_3dlut = false;
+
+#if PL_HAVE_LCMS
+
+    if (use_3dlut) {
+        struct pl_3dlut_profile src = {
+            .color = pass->cur_img.color,
+            .profile = image->profile,
+        };
+
+        struct pl_3dlut_profile dst = {
+            .color = target->color,
+            .profile = target->profile,
+        };
+
+        struct pl_3dlut_result res;
+        bool ok = pl_3dlut_update(sh, &src, &dst, &rr->lut3d_state, &res,
+                                  params->lut3d_params);
+        if (!ok) {
+            rr->disable_3dlut = true;
+            use_3dlut = false;
+            goto fallback;
+        }
+
+        // current -> 3DLUT in
+        pl_shader_color_map(sh, params->color_map_params, pass->cur_img.color,
+                            res.src_color, &rr->peak_detect_state, false);
+        // 3DLUT in -> 3DLUT out
+        pl_3dlut_apply(sh, &rr->lut3d_state);
+        // 3DLUT out -> target
+        pl_shader_color_map(sh, params->color_map_params, res.dst_color,
+                            target->color, &rr->peak_detect_state, false);
+    }
+
+fallback:
+
+#else // !PL_HAVE_LCMS
+
+    if (use_3dlut) {
+        PL_WARN(rr, "An ICC profile was set, but libplacebo is built without "
+                "support for LittleCMS! Disabling..");
+        rr->disable_3dlut = true;
+        use_3dlut = false;
+    }
+
+#endif
+
+    if (!use_3dlut) {
+        // current -> target
+        pl_shader_color_map(sh, params->color_map_params, pass->cur_img.color,
+                            target->color, &rr->peak_detect_state, false);
+    }
+
+    bool is_comp = pl_shader_is_compute(sh);
+    if (is_comp && !fbo->params.storable) {
+        bool ok = finalize_img(rr, &pass->cur_img, rr->fbofmt, &rr->output_fbo);
+        if (!ok) {
+            PL_ERR(rr, "Failed dispatching compute shader to intermediate FBO?");
+            return false;
+        }
+
+        sh = pass->cur_img.sh = pl_dispatch_begin(rr->dp);
+        pl_shader_sample_direct(sh, &(struct pl_sample_src) {
+            .tex = rr->output_fbo,
+        });
+    }
+
     pl_shader_encode_color(sh, &target->repr);
 
     // FIXME: Technically we should try dithering before bit shifting if we're
@@ -763,12 +926,6 @@ static bool pass_output_target(struct pl_renderer *rr, struct pass_state *pass,
             pl_shader_dither(sh, depth, &rr->dither_state, params->dither_params);
     }
 
-    bool is_comp = pl_shader_is_compute(sh);
-    if (is_comp && !fbo->params.storable) {
-        // TODO: force caching
-        abort();
-    }
-
     pl_assert(fbo->params.renderable);
     return pl_dispatch_finish(rr->dp, &pass->cur_img.sh, fbo,
                               &target->dst_rect, NULL);
@@ -817,6 +974,8 @@ bool pl_render_image(struct pl_renderer *rr, const struct pl_image *pimage,
     struct pl_image image = *pimage;
     struct pl_render_target target = *ptarget;
     fix_rects(&image, &target);
+    pl_color_space_infer(&image.color);
+    pl_color_space_infer(&target.color);
 
     // TODO: output caching
     pl_dispatch_reset_frame(rr->dp);
@@ -828,7 +987,7 @@ bool pl_render_image(struct pl_renderer *rr, const struct pl_image *pimage,
     if (!pass_scale_main(rr, &pass, &image, &target, params))
         goto error;
 
-    if (!pass_output_target(rr, &pass, &target, params))
+    if (!pass_output_target(rr, &pass, &image, &target, params))
         goto error;
 
     // If we don't have FBOs available, simulate the on-image overlays at
