@@ -568,11 +568,21 @@ ident_t sh_lut(struct pl_shader *sh, struct pl_shader_obj **obj,
         gpu ? gpu->limits.max_tex_3d_dim : 0,
     };
 
+    // Try picking the right number of dimensions for the texture LUT. This
+    // allows e.g. falling back to 2D textures if 1D textures are unsupported.
     for (int d = dims; d <= PL_ARRAY_SIZE(max_tex_dim); d++) {
-        if (size <= max_tex_dim[d - 1]) {
-            texdim = d;
-            break;
+        // For a given dimension to be compatible, all coordinates need to be
+        // within the maximum texture size for that dimension
+        for (int i = 0; i < d; i++) {
+            if (sizes[i] > max_tex_dim[d - 1])
+                goto next_dim;
         }
+
+        // All dimensions are compatible, so pick this texture dimension
+        texdim = d;
+        break;
+
+next_dim: ; // `continue` out of the inner loop
     }
 
     struct sh_lut_obj *lut = SH_OBJ(sh, obj, PL_SHADER_OBJ_LUT,
@@ -775,3 +785,27 @@ error:
     talloc_free(tmp);
     return ret;
 }
+
+const char *sh_bvec(const struct pl_shader *sh, int dims)
+{
+    static const char *bvecs[] = {
+        [1] = "bool",
+        [2] = "bvec2",
+        [3] = "bvec3",
+        [4] = "bvec4",
+    };
+
+    static const char *vecs[] = {
+        [1] = "float",
+        [2] = "vec2",
+        [3] = "vec3",
+        [4] = "vec4",
+    };
+
+    pl_assert(dims > 0 && dims < PL_ARRAY_SIZE(bvecs));
+    if (sh->gpu && sh->gpu->glsl.version >= 130) {
+        return bvecs[dims];
+    } else {
+        return vecs[dims];
+    }
+}
