@@ -68,25 +68,28 @@ void pl_shader_decode_color(struct pl_shader *sh, struct pl_color_repr *repr,
         // as per the BT.2020 specification, table 4. This is a non-linear
         // transformation because (constant) luminance receives non-equal
         // contributions from the three different channels.
-        GLSL("// constant luminance conversion                            \n"
-             "color.br = color.br * mix(vec2(1.5816, 0.9936),             \n"
-             "                          vec2(1.9404, 1.7184),             \n"
-             "                          lessThanEqual(color.br, vec2(0))) \n"
-             "           + color.gg;                                      \n"
+        GLSL("// constant luminance conversion                                \n"
+             "color.br = color.br * mix(vec2(1.5816, 0.9936),                 \n"
+             "                          vec2(1.9404, 1.7184),                 \n"
+             "                          %s(lessThanEqual(color.br, vec2(0)))) \n"
+             "           + color.gg;                                          \n",
+             sh_bvec(sh, 2));
         // Expand channels to camera-linear light. This shader currently just
         // assumes everything uses the BT.2020 12-bit gamma function, since the
         // difference between 10 and 12-bit is negligible for anything other
         // than 12-bit content.
-             "vec3 lin = mix(color.rgb * vec3(1.0/4.5),                        \n"
+        GLSL("vec3 lin = mix(color.rgb * vec3(1.0/4.5),                        \n"
              "                pow((color.rgb + vec3(0.0993))*vec3(1.0/1.0993), \n"
              "                    vec3(1.0/0.45)),                             \n"
-             "                lessThanEqual(vec3(0.08145), color.rgb));        \n"
+             "                %s(lessThanEqual(vec3(0.08145), color.rgb)));    \n",
+             sh_bvec(sh, 3));
         // Calculate the green channel from the expanded RYcB, and recompress to G'
         // The BT.2020 specification says Yc = 0.2627*R + 0.6780*G + 0.0593*B
-             "color.g = (lin.g - 0.2627*lin.r - 0.0593*lin.b)*1.0/0.6780; \n"
+        GLSL("color.g = (lin.g - 0.2627*lin.r - 0.0593*lin.b)*1.0/0.6780; \n"
              "color.g = mix(color.g * 4.5,                        \n"
              "              1.0993 * pow(color.g, 0.45) - 0.0993, \n"
-             "              0.0181 <= color.g);                   \n");
+             "              %s(0.0181 <= color.g));               \n",
+             sh_bvec(sh, 1));
     }
 
     if (repr->alpha == PL_ALPHA_INDEPENDENT) {
@@ -188,7 +191,8 @@ void pl_shader_linearize(struct pl_shader *sh, enum pl_color_transfer trc)
         GLSL("color.rgb = mix(color.rgb * vec3(1.0/12.92),               \n"
              "                pow((color.rgb + vec3(0.055))/vec3(1.055), \n"
              "                    vec3(2.4)),                            \n"
-             "                lessThan(vec3(0.04045), color.rgb));       \n");
+             "                %s(lessThan(vec3(0.04045), color.rgb)));   \n",
+             sh_bvec(sh, 3));
         break;
     case PL_COLOR_TRC_BT_1886:
         GLSL("color.rgb = pow(color.rgb, vec3(2.4));\n");
@@ -204,9 +208,10 @@ void pl_shader_linearize(struct pl_shader *sh, enum pl_color_transfer trc)
         GLSL("color.rgb = pow(color.rgb, vec3(2.8));\n");
         break;
     case PL_COLOR_TRC_PRO_PHOTO:
-        GLSL("color.rgb = mix(color.rgb * vec3(1.0/16.0),          \n"
-             "                pow(color.rgb, vec3(1.8)),           \n"
-             "                lessThan(vec3(0.03125), color.rgb)); \n");
+        GLSL("color.rgb = mix(color.rgb * vec3(1.0/16.0),              \n"
+             "                pow(color.rgb, vec3(1.8)),               \n"
+             "                %s(lessThan(vec3(0.03125), color.rgb))); \n",
+             sh_bvec(sh, 3));
         break;
     case PL_COLOR_TRC_PQ:
         GLSL("color.rgb = pow(color.rgb, vec3(1.0/%f));         \n"
@@ -222,15 +227,15 @@ void pl_shader_linearize(struct pl_shader *sh, enum pl_color_transfer trc)
         GLSL("color.rgb = mix(vec3(4.0) * color.rgb * color.rgb,         \n"
              "                exp((color.rgb - vec3(%f)) * vec3(1.0/%f)) \n"
              "                    + vec3(%f),                            \n"
-             "                lessThan(vec3(0.5), color.rgb));           \n",
-             HLG_C, HLG_A, HLG_B);
+             "                %s(lessThan(vec3(0.5), color.rgb)));       \n",
+             HLG_C, HLG_A, HLG_B, sh_bvec(sh, 3));
         break;
     case PL_COLOR_TRC_V_LOG:
         GLSL("color.rgb = mix((color.rgb - vec3(0.125)) * vec3(1.0/5.6), \n"
              "    pow(vec3(10.0), (color.rgb - vec3(%f)) * vec3(1.0/%f)) \n"
              "              - vec3(%f),                                  \n"
-             "    lessThanEqual(vec3(0.181), color.rgb));                \n",
-             VLOG_D, VLOG_C, VLOG_B);
+             "    %s(lessThanEqual(vec3(0.181), color.rgb)));            \n",
+             VLOG_D, VLOG_C, VLOG_B, sh_bvec(sh, 3));
         break;
     case PL_COLOR_TRC_S_LOG1:
         GLSL("color.rgb = pow(vec3(10.0), (color.rgb - vec3(%f)) * vec3(1.0/%f)) \n"
@@ -241,8 +246,9 @@ void pl_shader_linearize(struct pl_shader *sh, enum pl_color_transfer trc)
         GLSL("color.rgb = mix((color.rgb - vec3(%f)) * vec3(1.0/%f),      \n"
              "    (pow(vec3(10.0), (color.rgb - vec3(%f)) * vec3(1.0/%f)) \n"
              "              - vec3(%f)) * vec3(1.0/%f),                   \n"
-             "    lessThanEqual(vec3(%f), color.rgb));                    \n",
-             SLOG_Q, SLOG_P, SLOG_C, SLOG_A, SLOG_B, SLOG_K2, SLOG_Q);
+             "    %s(lessThanEqual(vec3(%f), color.rgb)));                \n",
+             SLOG_Q, SLOG_P, SLOG_C, SLOG_A, SLOG_B, SLOG_K2, sh_bvec(sh, 3),
+             SLOG_Q);
         break;
     default:
         abort();
@@ -262,10 +268,11 @@ void pl_shader_delinearize(struct pl_shader *sh, enum pl_color_transfer trc)
 
     switch (trc) {
     case PL_COLOR_TRC_SRGB:
-        GLSL("color.rgb = mix(color.rgb * vec3(12.92),                    \n"
-             "                vec3(1.055) * pow(color.rgb, vec3(1.0/2.4)) \n"
-             "                    - vec3(0.055),                          \n"
-             "                lessThanEqual(vec3(0.0031308), color.rgb)); \n");
+        GLSL("color.rgb = mix(color.rgb * vec3(12.92),                        \n"
+             "                vec3(1.055) * pow(color.rgb, vec3(1.0/2.4))     \n"
+             "                    - vec3(0.055),                              \n"
+             "                %s(lessThanEqual(vec3(0.0031308), color.rgb))); \n",
+             sh_bvec(sh, 3));
         break;
     case PL_COLOR_TRC_BT_1886:
         GLSL("color.rgb = pow(color.rgb, vec3(1.0/2.4));\n");
@@ -281,9 +288,10 @@ void pl_shader_delinearize(struct pl_shader *sh, enum pl_color_transfer trc)
         GLSL("color.rgb = pow(color.rgb, vec3(1.0/2.8));\n");
         break;
     case PL_COLOR_TRC_PRO_PHOTO:
-        GLSL("color.rgb = mix(color.rgb * vec3(16.0),                    \n"
-             "                pow(color.rgb, vec3(1.0/1.8)),             \n"
-             "                lessThanEqual(vec3(0.001953), color.rgb)); \n");
+        GLSL("color.rgb = mix(color.rgb * vec3(16.0),                        \n"
+             "                pow(color.rgb, vec3(1.0/1.8)),                 \n"
+             "                %s(lessThanEqual(vec3(0.001953), color.rgb))); \n",
+             sh_bvec(sh, 3));
         break;
     case PL_COLOR_TRC_PQ:
         GLSL("color.rgb *= vec3(1.0/%f);                         \n"
@@ -296,15 +304,15 @@ void pl_shader_delinearize(struct pl_shader *sh, enum pl_color_transfer trc)
     case PL_COLOR_TRC_HLG:
         GLSL("color.rgb = mix(vec3(0.5) * sqrt(color.rgb),                     \n"
              "                vec3(%f) * log(color.rgb - vec3(%f)) + vec3(%f), \n"
-             "                lessThan(vec3(1.0), color.rgb));                 \n",
-             HLG_A, HLG_B, HLG_C);
+             "                %s(lessThan(vec3(1.0), color.rgb)));             \n",
+             HLG_A, HLG_B, HLG_C, sh_bvec(sh, 3));
         break;
     case PL_COLOR_TRC_V_LOG:
-        GLSL("color.rgb = mix(vec3(5.6) * color.rgb + vec3(0.125),   \n"
-             "                vec3(%f) * log(color.rgb + vec3(%f))   \n"
-             "                    + vec3(%f),                        \n"
-             "                lessThanEqual(vec3(0.01), color.rgb)); \n",
-             VLOG_C / M_LN10, VLOG_B, VLOG_D);
+        GLSL("color.rgb = mix(vec3(5.6) * color.rgb + vec3(0.125),       \n"
+             "                vec3(%f) * log(color.rgb + vec3(%f))       \n"
+             "                    + vec3(%f),                            \n"
+             "                %s(lessThanEqual(vec3(0.01), color.rgb))); \n",
+             VLOG_C / M_LN10, VLOG_B, VLOG_D, sh_bvec(sh, 3));
         break;
     case PL_COLOR_TRC_S_LOG1:
         GLSL("color.rgb = vec3(%f) * log(color.rgb + vec3(%f)) + vec3(%f);\n",
@@ -314,8 +322,9 @@ void pl_shader_delinearize(struct pl_shader *sh, enum pl_color_transfer trc)
         GLSL("color.rgb = mix(vec3(%f) * color.rgb + vec3(%f),                \n"
              "                vec3(%f) * log(vec3(%f) * color.rgb + vec3(%f)) \n"
              "                    + vec3(%f),                                 \n"
-             "                lessThanEqual(vec3(0.0), color.rgb));           \n",
-             SLOG_P, SLOG_Q, SLOG_A / M_LN10, SLOG_K2, SLOG_B, SLOG_C);
+             "                %s(lessThanEqual(vec3(0.0), color.rgb)));       \n",
+             SLOG_P, SLOG_Q, SLOG_A / M_LN10, SLOG_K2, SLOG_B, SLOG_C,
+             sh_bvec(sh, 3));
         break;
     default:
         abort();
@@ -371,7 +380,7 @@ void pl_shader_unsigmoidize(struct pl_shader *sh,
 
 // Applies the OOTF / inverse OOTF
 static void pl_shader_ootf(struct pl_shader *sh, enum pl_color_light light,
-                           ident_t luma)
+                           ident_t luma, float peak)
 {
     if (!sh_require(sh, PL_SHADER_SIG_COLOR, 0, 0))
         return;
@@ -384,12 +393,14 @@ static void pl_shader_ootf(struct pl_shader *sh, enum pl_color_light light,
 
     switch (light)
     {
-    case PL_COLOR_LIGHT_SCENE_HLG:
-        // HLG OOTF from BT.2100, assuming a reference display with a
-        // peak of 1000 cd/m² -> gamma = 1.2
-        GLSL("color.rgb *= vec3(%f * pow(dot(%s, color.rgb), 0.2));\n",
-             (1000 / PL_COLOR_REF_WHITE) / pow(12, 1.2), luma);
+    case PL_COLOR_LIGHT_SCENE_HLG: {
+        // HLG OOTF from BT.2100, tuned to the indicated peak
+        float gamma = 1.2 + 0.42 * log10(peak * PL_COLOR_REF_WHITE / 1000.0);
+        gamma = PL_MAX(gamma, 1.0);
+        GLSL("color.rgb *= vec3(%f * pow(dot(%s, color.rgb), %f));\n",
+             peak / pow(12, gamma), luma, gamma - 1.0);
         break;
+    }
     case PL_COLOR_LIGHT_SCENE_709_1886:
         // This OOTF is defined by encoding the result as 709 and then decoding
         // it as 1886; although this is called 709_1886 we actually use the
@@ -397,8 +408,9 @@ static void pl_shader_ootf(struct pl_shader *sh, enum pl_color_light light,
         GLSL("color.rgb = mix(color.rgb * vec3(4.5),                    \n"
              "                vec3(1.0993) * pow(color.rgb, vec3(0.45)) \n"
              "                             - vec3(0.0993),              \n"
-             "                lessThan(vec3(0.0181), color.rgb));       \n"
-             "color.rgb = pow(color.rgb, vec3(2.4));                    \n");
+             "                %s(lessThan(vec3(0.0181), color.rgb)));   \n"
+             "color.rgb = pow(color.rgb, vec3(2.4));                    \n",
+             sh_bvec(sh, 3));
         break;
     case PL_COLOR_LIGHT_SCENE_1_2:
         GLSL("color.rgb = pow(color.rgb, vec3(1.2));\n");
@@ -409,7 +421,8 @@ static void pl_shader_ootf(struct pl_shader *sh, enum pl_color_light light,
 }
 
 static void pl_shader_inverse_ootf(struct pl_shader *sh,
-                                   enum pl_color_light light, ident_t luma)
+                                   enum pl_color_light light, ident_t luma,
+                                   float peak)
 {
     if (!light || light == PL_COLOR_LIGHT_DISPLAY)
         return;
@@ -419,18 +432,22 @@ static void pl_shader_inverse_ootf(struct pl_shader *sh,
 
     switch (light)
     {
-    case PL_COLOR_LIGHT_SCENE_HLG:
+    case PL_COLOR_LIGHT_SCENE_HLG: {
+        float gamma = 1.2 + 0.42 * log10(peak * PL_COLOR_REF_WHITE / 1000.0);
+        gamma = PL_MAX(gamma, 1.0);
         GLSL("color.rgb *= vec3(1.0/%f);                                \n"
              "color.rgb /= vec3(max(1e-6, pow(dot(%s, color.rgb),       \n"
-             "                                0.2/1.2)));               \n",
-             (1000 / PL_COLOR_REF_WHITE) / pow(12, 1.2), luma);
+             "                                %f)));                    \n",
+             peak / pow(12, gamma), luma, (gamma - 1.0) / gamma);
         break;
+    }
     case PL_COLOR_LIGHT_SCENE_709_1886:
         GLSL("color.rgb = pow(color.rgb, vec3(1.0/2.4));                         \n"
              "color.rgb = mix(color.rgb * vec3(1.0/4.5),                         \n"
              "                pow((color.rgb + vec3(0.0993)) * vec3(1.0/1.0993), \n"
              "                    vec3(1/0.45)),                                 \n"
-             "                lessThan(vec3(0.08145), color.rgb));               \n");
+             "                %s(lessThan(vec3(0.08145), color.rgb)));           \n",
+             sh_bvec(sh, 3));
         break;
     case PL_COLOR_LIGHT_SCENE_1_2:
         GLSL("color.rgb = pow(color.rgb, vec3(1.0/1.2));\n");
@@ -440,10 +457,6 @@ static void pl_shader_inverse_ootf(struct pl_shader *sh,
     }
 }
 
-// Average light level for SDR signals. This is equal to a signal level of 0.5
-// under a typical presentation gamma of about 2.0.
-static const float sdr_avg = 0.25;
-
 const struct pl_color_map_params pl_color_map_default_params = {
     .intent                  = PL_INTENT_RELATIVE_COLORIMETRIC,
     .tone_mapping_algo       = PL_TONE_MAPPING_HABLE,
@@ -470,7 +483,7 @@ static void hdr_update_peak(struct pl_shader *sh, struct pl_shader_obj **state,
     if (!state)
         return;
 
-    int frames = PL_DEF(params->peak_detect_frames, 20);
+    int frames = PL_DEF(params->peak_detect_frames, 63);
     if (frames < 1 || frames > 1000) {
         PL_ERR(sh, "Parameter peak_detect_frames must be >= 1 and <= 1000 "
                "(was %d).", frames);
@@ -564,7 +577,7 @@ static void hdr_update_peak(struct pl_shader *sh, struct pl_shader_obj **state,
          wg_sum, PL_COLOR_REF_WHITE);
 
     // Have one thread per work group update the global atomics. We use the
-    // work group average even for the global sum, to make the values slightly
+    // work group average even for the global max, to make the values slightly
     // more stable and smooth out tiny super-highlights.
     GLSL("memoryBarrierShared();                                            \n"
          "barrier();                                                        \n"
@@ -582,12 +595,11 @@ static void hdr_update_peak(struct pl_shader *sh, struct pl_shader_obj **state,
          "    float peak = float(%s) / (%f * float(%s)); \n"
          "    float avg  = float(%s) / (%f * float(%s)); \n"
          "    sig_peak   = max(1.0, peak);               \n"
-         "    sig_avg    = max(%f, avg);                 \n"
+         "    sig_avg    = max(0.25, avg);               \n"
          "}                                              \n",
          num.name,
          max_total.name, PL_COLOR_REF_WHITE, num.name,
-         avg_total.name, PL_COLOR_REF_WHITE, num.name,
-         sdr_avg);
+         avg_total.name, PL_COLOR_REF_WHITE, num.name);
 
     // Finally, to update the global state, we increment a counter per dispatch
     GLSL("memoryBarrierBuffer();                                                \n"
@@ -636,10 +648,10 @@ static void hdr_update_peak(struct pl_shader *sh, struct pl_shader_obj **state,
          max.name, avg.name);
 
     // Update the index and count
-    GLSL("    %s = next;                                                        \n"
-         "    %s = min(%s + 1, %d);                                             \n"
-         "    memoryBarrierBuffer();                                            \n"
-         "}                                                                     \n",
+    GLSL("    %s = next;             \n"
+         "    %s = min(%s + 1, %d);  \n"
+         "    memoryBarrierBuffer(); \n"
+         "}                          \n",
          idx.name,
          num.name, num.name, frames);
 }
@@ -672,16 +684,6 @@ static void pl_shader_tone_map(struct pl_shader *sh, struct pl_color_space src,
              dst.sig_peak, dst.sig_peak);
     }
 
-    // Desaturate the color using a coefficient dependent on the signal level
-    if (params->tone_mapping_desaturate > 0) {
-        GLSL("float luma = dot(%s, color.rgb);                      \n"
-             "float coeff = max(sig - 0.18, 1e-6) / max(sig, 1e-6); \n"
-             "coeff = pow(coeff, %f);                               \n"
-             "color.rgb = mix(color.rgb, vec3(luma), coeff);        \n"
-             "sig = mix(sig, luma, coeff);                          \n",
-             luma, 10.0 / params->tone_mapping_desaturate);
-    }
-
     // Store the original signal level for later re-use
     GLSL("float sig_orig = sig;\n");
 
@@ -691,6 +693,18 @@ static void pl_shader_tone_map(struct pl_shader *sh, struct pl_color_space src,
          "sig_peak *= slope;                    \n",
          dst.sig_avg);
 
+    // Desaturate the color using a coefficient dependent on the signal level.
+    // Do this after the brightness adjustment stage to prevent catastrophic
+    // saturation loss on overly bright HDR sources.
+    if (params->tone_mapping_desaturate > 0) {
+        GLSL("float luma = %f * dot(%s, color.rgb);                 \n"
+             "float coeff = max(sig - 0.18, 1e-6) / max(sig, 1e-6); \n"
+             "coeff = pow(coeff, %f);                               \n"
+             "color.rgb = mix(color.rgb, vec3(luma), coeff);        \n"
+             "sig = mix(sig, luma * slope, coeff);                  \n",
+             1.0 / dst.sig_peak, luma, 10.0 / params->tone_mapping_desaturate);
+    }
+
     float param = params->tone_mapping_param;
     switch (params->tone_mapping_algo) {
     case PL_TONE_MAPPING_CLIP:
@@ -698,14 +712,19 @@ static void pl_shader_tone_map(struct pl_shader *sh, struct pl_color_space src,
         break;
 
     case PL_TONE_MAPPING_MOBIUS:
-        GLSL("const float j = %f;                                           \n"
+        // Mobius isn't well-defined for sig_peak <= 1.0, but the limit of
+        // mobius as sig_peak -> 1.0 is a linear function, so we can just skip
+        // tone-mapping in this case
+        GLSL("if (sig_peak > 1.0 + 1e-6) {                                      \n"
+             "    const float j = %f;                                           \n"
              // solve for M(j) = j; M(sig_peak) = 1.0; M'(j) = 1.0
              // where M(x) = scale * (x+a)/(x+b)
-             "float a = -j*j * (sig_peak - 1.0) / (j*j - 2.0*j + sig_peak); \n"
-             "float b = (j*j - 2.0*j*sig_peak + sig_peak) /                 \n"
-             "          max(1e-6, sig_peak - 1.0);                          \n"
-             "float scale = (b*b + 2.0*b*j + j*j) / (b-a);                  \n"
-             "sig = sig > j ? (scale * (sig + a) / (sig + b)) : sig;        \n",
+             "    float a = -j*j * (sig_peak - 1.0) / (j*j - 2.0*j + sig_peak); \n"
+             "    float b = (j*j - 2.0*j*sig_peak + sig_peak) /                 \n"
+             "              max(1e-6, sig_peak - 1.0);                          \n"
+             "    float scale = (b*b + 2.0*b*j + j*j) / (b-a);                  \n"
+             "    sig = sig > j ? (scale * (sig + a) / (sig + b)) : sig;        \n"
+             "}                                                                 \n",
              PL_DEF(param, 0.3));
         break;
 
@@ -764,18 +783,8 @@ void pl_shader_color_map(struct pl_shader *sh,
     GLSL("{\n");
     params = PL_DEF(params, &pl_color_map_default_params);
 
-    // Defaults the primaries/transfer to sensible values. This isn't strictly
-    // necessary, but it avoids some redundant operations in the cases where
-    // src and dst are equal but one is set and the other is unknown
-    src.primaries = PL_DEF(src.primaries, PL_COLOR_PRIM_BT_709);
-    src.transfer = PL_DEF(src.transfer, PL_COLOR_TRC_GAMMA22);
-
-    // If the source light type is unknown, infer it from the transfer function.
-    if (!src.light) {
-        src.light = (src.transfer == PL_COLOR_TRC_HLG)
-            ? PL_COLOR_LIGHT_SCENE_HLG
-            : PL_COLOR_LIGHT_DISPLAY;
-    }
+    // Default the source color space to reasonable values
+    pl_color_space_infer(&src);
 
     // To be as conservative as possible, color mapping is disabled by default
     // except for special cases which are considered to be "sufficiently
@@ -795,40 +804,32 @@ void pl_shader_color_map(struct pl_shader *sh,
             dst.transfer = PL_COLOR_TRC_GAMMA22;
     }
 
-    // 99 times out of 100, this is what we want
-    dst.light = PL_DEF(dst.light, PL_COLOR_LIGHT_DISPLAY);
-
-    // Compute the highest encodable level
-    float src_range = pl_color_transfer_nominal_peak(src.transfer),
-          dst_range = pl_color_transfer_nominal_peak(dst.transfer);
-
-    // Default the src/dst peak information based on the encodable range. For
-    // the source peak, this is the safest possible value (no clipping). For
-    // the dest peak, this makes full use of the available dynamic range.
-    src.sig_peak = PL_DEF(src.sig_peak, src_range);
-    dst.sig_peak = PL_DEF(dst.sig_peak, dst_range);
-
-    // Defaults the signal average based on the SDR signal average.
-    // Note: For HDR, this assumes well-mastered HDR content.
-    src.sig_avg = PL_DEF(src.sig_avg, sdr_avg);
-
     // Defaults the dest average based on the source average, unless the source
-    // is HDR and the destination is not, in which case fall back to SDR avg.
+    // is HDR and the destination is not
     if (!dst.sig_avg) {
         bool src_hdr = pl_color_transfer_is_hdr(src.transfer);
         bool dst_hdr = pl_color_transfer_is_hdr(dst.transfer);
-        dst.sig_avg = src_hdr && !dst_hdr ? sdr_avg : src.sig_avg;
+        if (!(src_hdr && !dst_hdr))
+            dst.sig_avg = src.sig_avg;
     }
 
+    // Infer the remaining fields after making the above choices
+    pl_color_space_infer(&dst);
+
+    // HLG's OOTF is parametrized by the sig peak, so enable OOTF conversion
+    // if necessary, even if the output light is the same
+    bool need_ootf = src.light != dst.light;
+    if (src.transfer == PL_COLOR_TRC_HLG && src.sig_peak != dst.sig_peak)
+        need_ootf = true;
+
     // All operations from here on require linear light as a starting point,
     // so we linearize even if src.gamma == dst.gamma when one of the other
     // operations needs it
     bool need_linear = src.transfer != dst.transfer ||
                        src.primaries != dst.primaries ||
-                       src_range != dst_range ||
                        src.sig_peak > dst.sig_peak ||
                        src.sig_avg != dst.sig_avg ||
-                       src.light != dst.light;
+                       need_ootf;
 
     // Various operations need access to the src_luma and dst_luma respectively,
     // so just always make them available if we're doing anything at all
@@ -853,8 +854,8 @@ void pl_shader_color_map(struct pl_shader *sh,
         is_linear = true;
     }
 
-    if (src.light != dst.light)
-        pl_shader_ootf(sh, src.light, src_luma);
+    if (need_ootf)
+        pl_shader_ootf(sh, src.light, src_luma, src.sig_peak);
 
     // Adapt to the right colorspace (primaries) if necessary
     if (src.primaries != dst.primaries) {
@@ -867,9 +868,6 @@ void pl_shader_color_map(struct pl_shader *sh,
             .var = pl_var_mat3("cms_matrix"),
             .data = PL_TRANSPOSE_3X3(cms_mat.m),
         }));
-        // Since this can reduce the gamut, figure out by how much
-        for (int c = 0; c < 3; c++)
-            src.sig_peak = fmaxf(src.sig_peak, cms_mat.m[c][c]);
     }
 
     // Tone map to rescale the signal average/peak if needed
@@ -878,13 +876,17 @@ void pl_shader_color_map(struct pl_shader *sh,
 
     // Warn for remaining out-of-gamut colors if enabled
     if (params->gamut_warning) {
-        GLSL("if (any(greaterThan(color.rgb, vec3(1.01))) ||\n"
+        GLSL("if (any(greaterThan(color.rgb, vec3(%f + 0.01))) ||\n"
              "    any(lessThan(color.rgb, vec3(-0.01))))\n"
-             "    color.rgb = vec3(1.0) - color.rgb; // invert\n");
+             "    color.rgb = vec3(1.0) - color.rgb; // invert\n",
+             dst.sig_peak);
     }
 
-    if (src.light != dst.light)
-        pl_shader_inverse_ootf(sh, dst.light, dst_luma);
+    if (need_ootf)
+        pl_shader_inverse_ootf(sh, dst.light, dst_luma, dst.sig_peak);
+
+    if (params->hdr_simulation && !pl_color_transfer_is_hdr(dst.transfer))
+        GLSL("color.rgb *= vec3(%f);\n", 1.0 / dst.sig_peak);
 
     if (is_linear)
         pl_shader_delinearize(sh, dst.transfer);
@@ -1067,3 +1069,108 @@ const struct pl_dither_params pl_dither_default_params = {
     .method     = PL_DITHER_BLUE_NOISE,
     .temporal   = false, // commonly flickers on LCDs
 };
+
+#if PL_HAVE_LCMS
+
+#include "lcms.h"
+
+struct sh_3dlut_obj {
+    struct pl_context *ctx;
+    enum pl_rendering_intent intent;
+    struct pl_3dlut_profile src, dst;
+    struct pl_3dlut_result result;
+    struct pl_shader_obj *lut_obj;
+    bool updated; // to detect misuse of the API
+    bool ok;
+    ident_t lut;
+};
+
+static void sh_3dlut_uninit(const struct pl_gpu *gpu, void *ptr)
+{
+    struct sh_3dlut_obj *obj = ptr;
+    pl_shader_obj_destroy(&obj->lut_obj);
+    *obj = (struct sh_3dlut_obj) {0};
+}
+
+static void fill_3dlut(void *priv, float *data, int s_r, int s_g, int s_b)
+{
+    struct sh_3dlut_obj *obj = priv;
+    struct pl_context *ctx = obj->ctx;
+
+    obj->ok = pl_lcms_compute_lut(ctx, obj->intent, obj->src, obj->dst,
+                                  data, s_r, s_g, s_b, &obj->result);
+
+    if (!obj->ok)
+        pl_err(ctx, "Failed computing 3DLUT!");
+}
+
+static bool color_profile_eq(const struct pl_3dlut_profile *a,
+                             const struct pl_3dlut_profile *b)
+{
+    return pl_icc_profile_equal(&a->profile, &b->profile) &&
+           pl_color_space_equal(&a->color, &b->color);
+}
+
+bool pl_3dlut_update(struct pl_shader *sh,
+                     const struct pl_3dlut_profile *src,
+                     const struct pl_3dlut_profile *dst,
+                     struct pl_shader_obj **lut3d, struct pl_3dlut_result *out,
+                     const struct pl_3dlut_params *params)
+{
+    params = PL_DEF(params, &pl_3dlut_default_params);
+    size_t s_r = PL_DEF(params->size_r, 64),
+           s_g = PL_DEF(params->size_g, 64),
+           s_b = PL_DEF(params->size_b, 64);
+
+    struct sh_3dlut_obj *obj;
+    obj = SH_OBJ(sh, lut3d, PL_SHADER_OBJ_3DLUT,
+                 struct sh_3dlut_obj, sh_3dlut_uninit);
+    if (!obj)
+        return false;
+
+    bool changed = !color_profile_eq(&obj->src, src) ||
+                   !color_profile_eq(&obj->dst, dst) ||
+                   obj->intent != params->intent;
+
+    // Update the object, since we need this information from `fill_3dlut`
+    obj->ctx = sh->ctx;
+    obj->intent = params->intent;
+    obj->src = *src;
+    obj->dst = *dst;
+    obj->lut = sh_lut(sh, &obj->lut_obj, SH_LUT_LINEAR, s_r, s_g, s_b, 4,
+                      changed, obj, fill_3dlut);
+    if (!obj->lut || !obj->ok)
+        return false;
+
+    obj->updated = true;
+    *out = obj->result;
+    return true;
+}
+
+void pl_3dlut_apply(struct pl_shader *sh, struct pl_shader_obj **lut3d)
+{
+    if (!sh_require(sh, PL_SHADER_SIG_COLOR, 0, 0))
+        return;
+
+    struct sh_3dlut_obj *obj;
+    obj = SH_OBJ(sh, lut3d, PL_SHADER_OBJ_3DLUT,
+                 struct sh_3dlut_obj, sh_3dlut_uninit);
+    if (!obj || !obj->lut || !obj->updated || !obj->ok) {
+        PL_ERR(sh, "pl_shader_3dlut called without prior pl_3dlut_update?");
+        return;
+    }
+
+    GLSL("// pl_shader_3dlut\n");
+    GLSL("color.rgba = %s(color.rgb);\n", obj->lut);
+
+    obj->updated = false;
+}
+
+const struct pl_3dlut_params pl_3dlut_default_params = {
+    .intent = PL_INTENT_RELATIVE_COLORIMETRIC,
+    .size_r = 64,
+    .size_g = 64,
+    .size_b = 64,
+};
+
+#endif // PL_HAVE_LCMS
