@@ -25,15 +25,59 @@ const struct pl_deband_params pl_deband_default_params = {
     .grain      = 6.0,
 };
 
-void pl_shader_deband(struct pl_shader *sh, const struct pl_tex *pl_tex,
+// Helper function to compute the src/dst sizes and upscaling ratios
+static bool setup_src(struct pl_shader *sh, const struct pl_sample_src *src,
+                      ident_t *src_tex, ident_t *pos, ident_t *size, ident_t *pt,
+                      float *ratio_x, float *ratio_y, int *components,
+                      float *scale, bool resizeable)
+{
+    float src_w = pl_rect_w(src->rect);
+    float src_h = pl_rect_h(src->rect);
+    src_w = PL_DEF(src_w, src->tex->params.w);
+    src_h = PL_DEF(src_h, src->tex->params.h);
+
+    int out_w = PL_DEF(src->new_w, fabs(src_w));
+    int out_h = PL_DEF(src->new_h, fabs(src_h));
+
+    if (ratio_x)
+        *ratio_x = out_w / fabs(src_w);
+    if (ratio_y)
+        *ratio_y = out_h / fabs(src_h);
+    if (scale)
+        *scale = PL_DEF(src->scale, 1.0);
+
+    if (components) {
+        const struct pl_fmt *fmt = src->tex->params.format;
+        *components = PL_DEF(src->components, fmt->num_components);
+    }
+
+    if (resizeable)
+        out_w = out_h = 0;
+    if (!sh_require(sh, PL_SHADER_SIG_NONE, out_w, out_h))
+        return false;
+
+    struct pl_rect2df rect = {
+        .x0 = src->rect.x0,
+        .y0 = src->rect.y0,
+        .x1 = src->rect.x0 + src_w,
+        .y1 = src->rect.y0 + src_h,
+    };
+
+    *src_tex = sh_bind(sh, src->tex, "src_tex", &rect, pos, size, pt);
+    return true;
+}
+
+void pl_shader_deband(struct pl_shader *sh, const struct pl_sample_src *src,
                       const struct pl_deband_params *params)
 {
-    if (pl_tex->params.sample_mode != PL_TEX_SAMPLE_LINEAR) {
+    if (src->tex->params.sample_mode != PL_TEX_SAMPLE_LINEAR) {
         PL_ERR(sh, "Debanding requires sample_mode = PL_TEX_SAMPLE_LINEAR!");
         return;
     }
 
-    if (!sh_require(sh, PL_SHADER_SIG_NONE, 0, 0))
+    float scale;
+    ident_t tex, pos, pt;
+    if (!setup_src(sh, src, &tex, &pos, NULL, &pt, NULL, NULL, NULL, &scale, true))
         return;
 
     GLSL("vec4 color;\n");
@@ -41,11 +85,6 @@ void pl_shader_deband(struct pl_shader *sh, const struct pl_tex *pl_tex,
     GLSL("{\n");
     params = PL_DEF(params, &pl_deband_default_params);
 
-    ident_t tex, pos, pt;
-    tex = sh_bind(sh, pl_tex, "deband", NULL, &pos, NULL, &pt);
-    if (!tex)
-        return;
-
     ident_t prng, state;
     prng = sh_prng(sh, true, &state);
 
@@ -77,13 +116,15 @@ void pl_shader_deband(struct pl_shader *sh, const struct pl_tex *pl_tex,
     // For each iteration, compute the average at a given distance and
     // pick it instead of the color if the difference is below the threshold.
     for (int i = 1; i <= params->iterations; i++) {
-        GLSL("avg = %s(pos, %f, %s);                                \n"
-             "diff = abs(color - avg);                              \n"
-             "color = mix(avg, color, greaterThan(diff, vec4(%f))); \n",
+        GLSL("avg = %s(pos, %f, %s);                                    \n"
+             "diff = abs(color - avg);                                  \n"
+             "color = mix(avg, color, %s(greaterThan(diff, vec4(%f)))); \n",
              average, i * params->radius, state,
-             params->threshold / (1000 * i));
+             sh_bvec(sh, 4), params->threshold / (1000 * i * scale));
     }
 
+    GLSL("color *= vec4(%f);\n", scale);
+
     // Add some random noise to smooth out residual differences
     if (params->grain > 0) {
         GLSL("vec3 noise = vec3(%s, %s, %s);         \n"
@@ -94,55 +135,16 @@ void pl_shader_deband(struct pl_shader *sh, const struct pl_tex *pl_tex,
     GLSL("}\n");
 }
 
-// Helper function to compute the src/dst sizes and upscaling ratios
-static bool setup_src(struct pl_shader *sh, const struct pl_sample_src *src,
-                      ident_t *src_tex, ident_t *pos, ident_t *size, ident_t *pt,
-                      float *ratio_x, float *ratio_y, int *components,
-                      bool resizeable)
-{
-    float src_w = pl_rect_w(src->rect);
-    float src_h = pl_rect_h(src->rect);
-    src_w = PL_DEF(src_w, src->tex->params.w);
-    src_h = PL_DEF(src_h, src->tex->params.h);
-
-    int out_w = PL_DEF(src->new_w, fabs(src_w));
-    int out_h = PL_DEF(src->new_h, fabs(src_h));
-
-    if (ratio_x)
-        *ratio_x = out_w / fabs(src_w);
-    if (ratio_y)
-        *ratio_y = out_h / fabs(src_h);
-
-    if (components) {
-        const struct pl_fmt *fmt = src->tex->params.format;
-        *components = PL_DEF(src->components, fmt->num_components);
-    }
-
-    if (resizeable)
-        out_w = out_h = 0;
-    if (!sh_require(sh, PL_SHADER_SIG_NONE, out_w, out_h))
-        return false;
-
-    struct pl_rect2df rect = {
-        .x0 = src->rect.x0,
-        .y0 = src->rect.y0,
-        .x1 = src->rect.x0 + src_w,
-        .y1 = src->rect.y0 + src_h,
-    };
-
-    *src_tex = sh_bind(sh, src->tex, "src_tex", &rect, pos, size, pt);
-    return true;
-}
-
 bool pl_shader_sample_direct(struct pl_shader *sh, const struct pl_sample_src *src)
 {
+    float scale;
     ident_t tex, pos;
-    if (!setup_src(sh, src, &tex, &pos, NULL, NULL, NULL, NULL, NULL, true))
+    if (!setup_src(sh, src, &tex, &pos, NULL, NULL, NULL, NULL, NULL, &scale, true))
         return false;
 
-    GLSL("// pl_shader_sample_direct    \n"
-         "vec4 color = texture(%s, %s); \n",
-         tex, pos);
+    GLSL("// pl_shader_sample_direct               \n"
+         "vec4 color = vec4(%f) * texture(%s, %s); \n",
+         scale, tex, pos);
     return true;
 }
 
@@ -173,8 +175,8 @@ bool pl_shader_sample_bicubic(struct pl_shader *sh, const struct pl_sample_src *
     }
 
     ident_t tex, pos, size, pt;
-    float rx, ry;
-    if (!setup_src(sh, src, &tex, &pos, &size, &pt, &rx, &ry, NULL, true))
+    float rx, ry, scale;
+    if (!setup_src(sh, src, &tex, &pos, &size, &pt, &rx, &ry, NULL, &scale, true))
         return false;
 
     if (rx < 1 || ry < 1) {
@@ -206,9 +208,9 @@ bool pl_shader_sample_bicubic(struct pl_shader *sh, const struct pl_sample_src *
          "vec4 bg = texture(%s, pos + cdelta.zw);       \n"
          "vec4 aa = mix(bg, br, parmy.b);               \n"
          // x-interpolation
-         "color = mix(aa, ab, parmx.b);                 \n"
+         "color = vec4(%f) * mix(aa, ab, parmx.b);      \n"
          "}                                             \n",
-         tex, tex, tex, tex);
+         tex, tex, tex, tex, scale);
     return true;
 }
 
@@ -316,9 +318,9 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
     }
 
     int comps;
-    float rx, ry;
+    float rx, ry, scale;
     ident_t src_tex, pos, size, pt;
-    if (!setup_src(sh, src, &src_tex, &pos, &size, &pt, &rx, &ry, &comps, false))
+    if (!setup_src(sh, src, &src_tex, &pos, &size, &pt, &rx, &ry, &comps, &scale, false))
         return false;
 
     struct sh_sampler_obj *obj;
@@ -481,8 +483,9 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
         }
     }
 
-    GLSL("color = color / vec4(wsum); \n"
-         "}                           \n");
+    GLSL("color = vec4(%f / wsum) * color; \n"
+         "}                                \n",
+         scale);
     return true;
 }
 
@@ -537,10 +540,10 @@ bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
     }
 
     int comps;
-    float ratio[2];
+    float ratio[2], scale;
     ident_t src_tex, pos, size, pt;
     if (!setup_src(sh, &srcfix, &src_tex, &pos, &size, &pt, &ratio[1], &ratio[0],
-                   &comps, false))
+                   &comps, &scale, false))
     {
         return false;
     }
@@ -654,6 +657,7 @@ bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
              params->antiring);
     }
 
+    GLSL("color *= vec4(%f);\n", scale);
     GLSL("}\n");
     return true;
 }
