@@ -1,5 +1,5 @@
 #include "tests.h"
-#include "time.h"
+#include <sys/time.h>
 
 #define TEX_SIZE 2048
 #define CUBE_SIZE 64
@@ -98,12 +98,7 @@ static void run_bench(const struct pl_gpu *gpu, struct pl_dispatch *dp,
     bool ok = pl_tex_download(gpu, &(struct pl_tex_transfer_params) {
         .tex        = fbo.tex,
         .buf        = fbo.buf,
-        // Transfer a single pixel:
-        .stride_w   = 1,
-        .rc         = {
-            0, 0, 0,
-            1, 1, 1
-        },
+        .rc         = { .x1 = 1, .y1 = 1 }, // 1x1 pixels
     });
     REQUIRE(ok);
 }
@@ -122,18 +117,20 @@ static void benchmark(const struct pl_gpu *gpu, const char *name, bench_fn bench
     while (pl_buf_poll(gpu, fbos[0].buf, 1000000000)); // 1 s
 
     // Perform the actual benchmark
-    clock_t start = clock(), stop = {0};
+    struct timeval start = {0}, stop = {0};
     unsigned long frames = 0;
     int index = 0;
 
+    gettimeofday(&start, NULL);
     do {
         frames++;
         run_bench(gpu, dp, &state, src, fbos[index++], bench);
         index %= NUM_FBOS;
-        stop = clock();
-    } while (stop - start < BENCH_DUR * CLOCKS_PER_SEC);
+        gettimeofday(&stop, NULL);
+    } while (stop.tv_sec - start.tv_sec < BENCH_DUR);
 
-    float secs = (float) (stop - start) / CLOCKS_PER_SEC;
+    float secs = (float) (stop.tv_sec - start.tv_sec) +
+                 1e-6 * (stop.tv_usec - start.tv_usec);
     printf("'%s':\t%4lu frames in %1.6f seconds => %2.6f ms/frame (%5.2f FPS)\n",
           name, frames, secs, 1000 * secs / frames, frames / secs);
 
@@ -162,13 +159,14 @@ static void bench_bt2020c(struct pl_shader *sh, struct pl_shader_obj **state,
 static void bench_deband(struct pl_shader *sh, struct pl_shader_obj **state,
                          const struct pl_tex *src)
 {
-    pl_shader_deband(sh, src, NULL);
+    pl_shader_deband(sh, &(struct pl_sample_src) { .tex = src }, NULL);
 }
 
 static void bench_deband_heavy(struct pl_shader *sh, struct pl_shader_obj **state,
                                const struct pl_tex *src)
 {
-    pl_shader_deband(sh, src, &(struct pl_deband_params) {
+    pl_shader_deband(sh, &(struct pl_sample_src) { .tex = src },
+        &(struct pl_deband_params) {
         .iterations = 4,
         .threshold  = 4.0,
         .radius     = 4.0,
