@@ -22,6 +22,38 @@
 
 const struct pl_vk_inst_params pl_vk_inst_default_params = {0};
 
+struct vk_ext_fun {
+    const char *name;
+    size_t offset;
+};
+
+struct vk_ext {
+    const char *name;
+    struct vk_ext_fun *funs;
+};
+
+#define VK_DEV_FUN(N)                       \
+    { .name = #N,                           \
+      .offset = offsetof(struct vk_ctx, N), \
+    }
+
+// Table of vulkan instance extensions
+// TODO: add support for loading function pointers
+static const char *vk_instance_extensions[] = {
+    VK_EXT_SWAPCHAIN_COLOR_SPACE_EXTENSION_NAME,
+};
+
+// Table of vulkan device extensions and functions they load
+static const struct vk_ext vk_device_extensions[] = {
+    {
+        .name = VK_KHR_PUSH_DESCRIPTOR_EXTENSION_NAME,
+        .funs = (struct vk_ext_fun[]) {
+            VK_DEV_FUN(vkCmdPushDescriptorSetKHR),
+            {0},
+        },
+    }
+};
+
 void pl_vk_inst_destroy(const struct pl_vk_inst **inst_ptr)
 {
     const struct pl_vk_inst *inst = *inst_ptr;
@@ -30,8 +62,8 @@ void pl_vk_inst_destroy(const struct pl_vk_inst **inst_ptr)
 
     VkDebugReportCallbackEXT debug = (VkDebugReportCallbackEXT) inst->priv;
     if (debug) {
-        VK_LOAD_PFN(inst->instance, vkDestroyDebugReportCallbackEXT)
-        pfn_vkDestroyDebugReportCallbackEXT(inst->instance, debug, VK_ALLOC);
+        VK_LOAD_FUN(inst->instance, vkDestroyDebugReportCallbackEXT)
+        vkDestroyDebugReportCallbackEXT(inst->instance, debug, VK_ALLOC);
     }
 
     vkDestroyInstance(inst->instance, VK_ALLOC);
@@ -61,7 +93,7 @@ static VkBool32 vk_dbg_callback(VkDebugReportFlagsEXT flags,
     // The return value of this function determines whether the call will
     // be explicitly aborted (to prevent GPU errors) or not. In this case,
     // we generally want this to be on for the errors.
-    return (flags & VK_DEBUG_REPORT_ERROR_BIT_EXT);
+    return !!(flags & VK_DEBUG_REPORT_ERROR_BIT_EXT);
 }
 
 const struct pl_vk_inst *pl_vk_inst_create(struct pl_context *ctx,
@@ -69,6 +101,7 @@ const struct pl_vk_inst *pl_vk_inst_create(struct pl_context *ctx,
 {
     void *tmp = talloc_new(NULL);
     params = PL_DEF(params, &pl_vk_inst_default_params);
+    VkInstance inst = NULL;
 
     const char **exts = NULL;
     int num_exts = 0;
@@ -77,10 +110,41 @@ const struct pl_vk_inst *pl_vk_inst_create(struct pl_context *ctx,
         .sType = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO,
     };
 
+    // Enumerate all supported extensions
+    uint32_t num_exts_avail = 0;
+    vkEnumerateInstanceExtensionProperties(NULL, &num_exts_avail, NULL);
+    VkExtensionProperties *exts_avail = talloc_array(tmp, VkExtensionProperties, num_exts_avail);
+    vkEnumerateInstanceExtensionProperties(NULL, &num_exts_avail, exts_avail);
+
+    // Add mandatory extensions
+    TARRAY_APPEND(tmp, exts, num_exts, VK_KHR_GET_PHYSICAL_DEVICE_PROPERTIES_2_EXTENSION_NAME);
+
+    // Add optional extensions
+    for (int i = 0; i < PL_ARRAY_SIZE(vk_instance_extensions); i++) {
+        const char *ext = vk_instance_extensions[i];
+        for (int n = 0; n < num_exts_avail; n++) {
+            if (strcmp(ext, exts_avail[n].extensionName) == 0) {
+                TARRAY_APPEND(tmp, exts, num_exts, ext);
+                break;
+            }
+        }
+    }
+
     // Add extra user extensions
     for (int i = 0; i < params->num_extensions; i++)
         TARRAY_APPEND(tmp, exts, num_exts, params->extensions[i]);
 
+    // Add extra optional user extensions
+    for (int i = 0; i < params->num_opt_extensions; i++) {
+        const char *ext = params->opt_extensions[i];
+        for (int n = 0; n < num_exts_avail; n++) {
+            if (strcmp(ext, exts_avail[n].extensionName) == 0) {
+                TARRAY_APPEND(tmp, exts, num_exts, ext);
+                break;
+            }
+        }
+    }
+
     if (params->debug) {
         pl_info(ctx, "Enabling vulkan debug layers");
         // Enables the LunarG standard validation layer, which
@@ -98,7 +162,6 @@ const struct pl_vk_inst *pl_vk_inst_create(struct pl_context *ctx,
     info.ppEnabledExtensionNames = exts;
     info.enabledExtensionCount = num_exts;
 
-    VkInstance inst = NULL;
     VkResult res = vkCreateInstance(&info, VK_ALLOC, &inst);
     if (res != VK_SUCCESS) {
         pl_fatal(ctx, "Failed creating instance: %s", vk_res_str(res));
@@ -122,19 +185,24 @@ const struct pl_vk_inst *pl_vk_inst_create(struct pl_context *ctx,
         // Since this is not part of the core spec, we need to load it. This
         // can't fail because we've already successfully created an instance
         // with this extension enabled.
-        VK_LOAD_PFN(inst, vkCreateDebugReportCallbackEXT)
-        pfn_vkCreateDebugReportCallbackEXT(inst, &dinfo, VK_ALLOC, &debug);
+        VK_LOAD_FUN(inst, vkCreateDebugReportCallbackEXT)
+        vkCreateDebugReportCallbackEXT(inst, &dinfo, VK_ALLOC, &debug);
     }
 
-    talloc_free(tmp);
-    return talloc_struct(NULL, struct pl_vk_inst, {
+    struct pl_vk_inst *pl_vk = talloc_struct(NULL, struct pl_vk_inst, {
         .instance = inst,
         .priv = (uint64_t) debug,
+        .extensions = exts,
+        .num_extensions = num_exts,
     });
 
+    pl_vk->extensions = talloc_steal(pl_vk, pl_vk->extensions);
+    talloc_free(tmp);
+    return pl_vk;
+
 error:
     pl_fatal(ctx, "Failed initializing vulkan instance");
-    pl_vk_inst_destroy((const struct pl_vk_inst **) &inst);
+    vkDestroyInstance(inst, VK_ALLOC);
     talloc_free(tmp);
     return NULL;
 }
@@ -365,14 +433,47 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
     add_qinfo(tmp, &qinfos, &num_qinfos, qfs, idx_comp, params->queue_count);
     add_qinfo(tmp, &qinfos, &num_qinfos, qfs, idx_tf, params->queue_count);
 
-    const char **exts = NULL;
-    int num_exts = 0;
+    // Enumerate all supported extensions
+    uint32_t num_exts_avail = 0;
+    VK(vkEnumerateDeviceExtensionProperties(vk->physd, NULL, &num_exts_avail, NULL));
+    VkExtensionProperties *exts_avail = talloc_array(tmp, VkExtensionProperties, num_exts_avail);
+    VK(vkEnumerateDeviceExtensionProperties(vk->physd, NULL, &num_exts_avail, exts_avail));
+
+    // Add all extensions we need
+    const char ***exts = &vk->exts;
+    int *num_exts = &vk->num_exts;
     if (params->surface)
-        TARRAY_APPEND(tmp, exts, num_exts, VK_KHR_SWAPCHAIN_EXTENSION_NAME);
+        TARRAY_APPEND(vk, *exts, *num_exts, VK_KHR_SWAPCHAIN_EXTENSION_NAME);
+
+    // Add all extensions we can benefit from, and remember their functions
+    const struct vk_ext_fun **ext_funs = NULL;
+    int num_ext_funs = 0;
+    for (int i = 0; i < PL_ARRAY_SIZE(vk_device_extensions); i++) {
+        const struct vk_ext *ext = &vk_device_extensions[i];
+        for (int n = 0; n < num_exts_avail; n++) {
+            if (strcmp(ext->name, exts_avail[n].extensionName) == 0) {
+                TARRAY_APPEND(vk, *exts, *num_exts, ext->name);
+                for (const struct vk_ext_fun *f = ext->funs; f->name; f++)
+                    TARRAY_APPEND(tmp, ext_funs, num_ext_funs, f);
+                break;
+            }
+        }
+    }
 
     // Add extra user extensions
     for (int i = 0; i < params->num_extensions; i++)
-        TARRAY_APPEND(tmp, exts, num_exts, params->extensions[i]);
+        TARRAY_APPEND(vk, *exts, *num_exts, params->extensions[i]);
+
+    // Add optional extra user extensions
+    for (int i = 0; i < params->num_opt_extensions; i++) {
+        const char *ext = params->opt_extensions[i];
+        for (int n = 0; n < num_exts_avail; n++) {
+            if (strcmp(ext, exts_avail[n].extensionName) == 0) {
+                TARRAY_APPEND(vk, *exts, *num_exts, ext);
+                break;
+            }
+        }
+    }
 
     // Enable all features that we might need (whitelisted)
     vkGetPhysicalDeviceFeatures(vk->physd, &vk->features);
@@ -387,17 +488,24 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
         .sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO,
         .pQueueCreateInfos = qinfos,
         .queueCreateInfoCount = num_qinfos,
-        .ppEnabledExtensionNames = exts,
-        .enabledExtensionCount = num_exts,
+        .ppEnabledExtensionNames = *exts,
+        .enabledExtensionCount = *num_exts,
         .pEnabledFeatures = &vk->features,
     };
 
-    PL_INFO(vk, "Creating vulkan device%s", num_exts ? " with extensions:" : "");
-    for (int i = 0; i < num_exts; i++)
-        PL_INFO(vk, "    %s", exts[i]);
+    PL_INFO(vk, "Creating vulkan device%s", *num_exts ? " with extensions:" : "");
+    for (int i = 0; i < *num_exts; i++)
+        PL_INFO(vk, "    %s", (*exts)[i]);
 
     VK(vkCreateDevice(vk->physd, &dinfo, VK_ALLOC, &vk->dev));
 
+    // Load all of the functions from the extensions we enabled
+    for (int i = 0; i < num_ext_funs; i++) {
+        const struct vk_ext_fun *fun = ext_funs[i];
+        void *pfn = (void *) ((uintptr_t) vk + (ptrdiff_t) fun->offset);
+        *((PFN_vkVoidFunction *) pfn) = vkGetDeviceProcAddr(vk->dev, fun->name);
+    }
+
     // Create the command pools and memory allocator
     for (int i = 0; i < num_qinfos; i++) {
         int qf = qinfos[i].queueFamilyIndex;
@@ -477,6 +585,18 @@ const struct pl_vulkan *pl_vulkan_create(struct pl_context *ctx,
     pl_vk->instance = vk->inst;
     pl_vk->phys_device = vk->physd;
     pl_vk->device = vk->dev;
+    pl_vk->extensions = vk->exts;
+    pl_vk->num_extensions = vk->num_exts;
+    pl_vk->num_queues = vk->num_pools;
+    pl_vk->queues = talloc_array(pl_vk, struct pl_vulkan_queue, vk->num_pools);
+    for (int i = 0; i < vk->num_pools; i++) {
+        struct pl_vulkan_queue *queues = (struct pl_vulkan_queue *) pl_vk->queues;
+        queues[i] = (struct pl_vulkan_queue) {
+            .index = vk->pools[i]->qf,
+            .count = vk->pools[i]->num_queues,
+        };
+    }
+
     return pl_vk;
 
 error:
