@@ -35,6 +35,10 @@ struct pl_vk {
     struct vk_malloc *alloc;
     struct spirv_compiler *spirv;
 
+    // Some additional cached device limits
+    uint32_t max_push_descriptors;
+    size_t min_texel_alignment;
+
     // This is a pl_dispatch used (on ourselves!) for the purposes of
     // dispatching compute shaders for performing various emulation tasks
     // (e.g. partial clears, blits or emulated texture transfers).
@@ -128,6 +132,9 @@ static void vk_setup_formats(struct pl_gpu *gpu)
 {
     struct vk_ctx *vk = pl_vk_get(gpu);
 
+    // Texture format emulation requires at least support for texel buffers
+    bool has_emu = (gpu->caps & PL_GPU_CAP_COMPUTE) && gpu->limits.max_buffer_texels;
+
     for (const struct vk_format *vk_fmt = vk_formats; vk_fmt->ifmt; vk_fmt++) {
         VkFormatProperties prop;
         vkGetPhysicalDeviceFormatProperties(vk->physd, vk_fmt->ifmt, &prop);
@@ -161,7 +168,7 @@ static void vk_setup_formats(struct pl_gpu *gpu)
 
         // For the texture capabilities, try falling back to the emulation
         // format if this format is wholly unsupported.
-        if (!prop.optimalTilingFeatures && vk_fmt->emufmt) {
+        if (has_emu && !prop.optimalTilingFeatures && vk_fmt->emufmt) {
             fmt->emulated = true;
             vkGetPhysicalDeviceFormatProperties(vk->physd, vk_fmt->emufmt, &prop);
         }
@@ -237,6 +244,21 @@ const struct pl_gpu *pl_gpu_create_vk(struct vk_ctx *vk)
         .align_tex_xfer_offset = vk->limits.optimalBufferCopyOffsetAlignment,
     };
 
+    if (vk->vkCmdPushDescriptorSetKHR) {
+        VkPhysicalDevicePushDescriptorPropertiesKHR pushd = {
+            .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PUSH_DESCRIPTOR_PROPERTIES_KHR,
+        };
+
+        VkPhysicalDeviceProperties2KHR props = {
+            .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROPERTIES_2_KHR,
+            .pNext = &pushd,
+        };
+
+        VK_LOAD_FUN(vk->inst, vkGetPhysicalDeviceProperties2KHR);
+        vkGetPhysicalDeviceProperties2KHR(vk->physd, &props);
+        p->max_push_descriptors = pushd.maxPushDescriptors;
+    }
+
     if (vk->pool_compute) {
         gpu->caps |= PL_GPU_CAP_COMPUTE;
         gpu->limits.max_shmem_size = vk->limits.maxComputeSharedMemorySize;
@@ -259,6 +281,14 @@ const struct pl_gpu *pl_gpu_create_vk(struct vk_ctx *vk)
 
     vk_setup_formats(gpu);
 
+    // Compute the correct minimum texture alignment
+    p->min_texel_alignment = 1;
+    for (int i = 0; i < gpu->num_formats; i++) {
+        size_t texel_size = gpu->formats[i]->texel_size;
+        p->min_texel_alignment = pl_lcm(p->min_texel_alignment, texel_size);
+    }
+    PL_DEBUG(gpu, "Minimum texel alignment: %zu", p->min_texel_alignment);
+
     // Create the dispatch last, after any setup of `gpu` is done
     p->dp = pl_dispatch_create(vk->ctx, gpu);
     pl_gpu_print_info(gpu, PL_LOG_INFO);
@@ -308,6 +338,7 @@ static VkResult vk_create_render_pass(VkDevice dev, const struct pl_fmt *fmt,
 
 // For pl_tex.priv
 struct pl_tex_vk {
+    bool held;
     bool external_img;
     bool may_invalidate;
     enum queue_type transfer_queue;
@@ -339,6 +370,9 @@ struct pl_tex_vk {
 void pl_tex_vk_external_dep(const struct pl_gpu *gpu, const struct pl_tex *tex,
                             VkSemaphore external_dep)
 {
+    if (!external_dep)
+        return;
+
     struct pl_tex_vk *tex_vk = tex->priv;
     TARRAY_APPEND(tex_vk, tex_vk->ext_deps, tex_vk->num_ext_deps, external_dep);
 }
@@ -351,6 +385,7 @@ static void tex_barrier(const struct pl_gpu *gpu, struct vk_cmd *cmd,
 {
     struct vk_ctx *vk = pl_vk_get(gpu);
     struct pl_tex_vk *tex_vk = tex->priv;
+    assert(!tex_vk->held);
 
     for (int i = 0; i < tex_vk->num_ext_deps; i++)
         vk_cmd_dep(cmd, tex_vk->ext_deps[i], stage);
@@ -593,6 +628,20 @@ static const struct pl_tex *vk_tex_create(const struct pl_gpu *gpu,
             goto error;
         }
 
+        // Statically check to see if we'd even be able to upload it at all
+        // and refuse right away if not. In theory, uploading can still fail
+        // based on the size of pl_tex_transfer_params.stride_w, but for now
+        // this should be enough.
+        uint64_t texels = params->w * params->h * params->d *
+                          params->format->num_components;
+
+        if (texels > gpu->limits.max_buffer_texels) {
+            PL_ERR(gpu, "Failed creating texture with emulated texture format: "
+                   "texture dimensions exceed maximum texel buffer size! Try "
+                   "again with a different (non-emulated) format?");
+            goto error;
+        }
+
         // Our format emulation requires storage image support. In order to
         // make a bunch of checks happy, just mark it off as storable (and also
         // enable VK_IMAGE_USAGE_STORAGE_BIT, which we do below)
@@ -799,44 +848,48 @@ static void vk_tex_blit(const struct pl_gpu *gpu,
     tex_signal(gpu, cmd, dst, VK_PIPELINE_STAGE_TRANSFER_BIT);
 }
 
-const struct pl_tex *pl_vk_wrap_swimg(const struct pl_gpu *gpu, VkImage vkimg,
-                                      VkSwapchainCreateInfoKHR info)
+const struct pl_tex *pl_vulkan_wrap(const struct pl_gpu *gpu,
+                                    VkImage image, int w, int h, int d,
+                                    VkFormat imageFormat,
+                                    VkImageUsageFlags imageUsage)
 {
     struct pl_tex *tex = NULL;
 
     const struct pl_fmt *format = NULL;
     for (int i = 0; i < gpu->num_formats; i++) {
         const struct vk_format *fmt = gpu->formats[i]->priv;
-        if (fmt->ifmt == info.imageFormat) {
+        if (fmt->ifmt == imageFormat) {
             format = gpu->formats[i];
             break;
         }
     }
 
     if (!format) {
-        PL_ERR(gpu, "Could not find pl_fmt suitable for wrapped swchain image "
-               "with surface format 0x%x\n", (unsigned) info.imageFormat);
+        PL_ERR(gpu, "Could not find pl_fmt suitable for wrapped image "
+               "with VkFormat 0x%x\n", (unsigned) imageFormat);
         goto error;
     }
 
     tex = talloc_zero(NULL, struct pl_tex);
     tex->params = (struct pl_tex_params) {
         .format = format,
-        .w = info.imageExtent.width,
-        .h = info.imageExtent.height,
-        .sampleable  = !!(info.imageUsage & VK_IMAGE_USAGE_SAMPLED_BIT),
-        .renderable  = !!(info.imageUsage & VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT),
-        .storable    = !!(info.imageUsage & VK_IMAGE_USAGE_STORAGE_BIT),
-        .blit_src    = !!(info.imageUsage & VK_IMAGE_USAGE_TRANSFER_SRC_BIT),
-        .blit_dst    = !!(info.imageUsage & VK_IMAGE_USAGE_TRANSFER_DST_BIT),
-        .host_writable = !!(info.imageUsage & VK_IMAGE_USAGE_TRANSFER_DST_BIT),
-        .host_readable = !!(info.imageUsage & VK_IMAGE_USAGE_TRANSFER_SRC_BIT),
+        .w = w,
+        .h = h,
+        .d = d,
+        .sampleable  = !!(imageUsage & VK_IMAGE_USAGE_SAMPLED_BIT),
+        .renderable  = !!(imageUsage & VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT),
+        .storable    = !!(imageUsage & VK_IMAGE_USAGE_STORAGE_BIT),
+        .blit_src    = !!(imageUsage & VK_IMAGE_USAGE_TRANSFER_SRC_BIT),
+        .blit_dst    = !!(imageUsage & VK_IMAGE_USAGE_TRANSFER_DST_BIT),
+        .host_writable = !!(imageUsage & VK_IMAGE_USAGE_TRANSFER_DST_BIT),
+        .host_readable = !!(imageUsage & VK_IMAGE_USAGE_TRANSFER_SRC_BIT),
     };
 
     struct pl_tex_vk *tex_vk = tex->priv = talloc_zero(tex, struct pl_tex_vk);
     tex_vk->type = VK_IMAGE_TYPE_2D;
     tex_vk->external_img = true;
-    tex_vk->img = vkimg;
+    tex_vk->held = true;
+    tex_vk->img = image;
 
     if (!vk_init_image(gpu, tex))
         goto error;
@@ -848,6 +901,41 @@ error:
     return NULL;
 }
 
+void pl_vulkan_hold(const struct pl_gpu *gpu, const struct pl_tex *tex,
+                    VkImageLayout layout, VkAccessFlags access,
+                    VkSemaphore sem_out)
+{
+    struct pl_tex_vk *tex_vk = tex->priv;
+    pl_assert(tex_vk->external_img);
+    pl_assert(!tex_vk->held);
+    pl_assert(sem_out);
+
+    struct vk_cmd *cmd = vk_require_cmd(gpu, GRAPHICS);
+    if (!cmd) {
+        PL_ERR(gpu, "Failed holding external image!");
+        return;
+    }
+
+    tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
+                access, layout);
+    vk_cmd_sig(cmd, sem_out);
+    pl_gpu_flush(gpu);
+    tex_vk->held = true;
+}
+
+void pl_vulkan_release(const struct pl_gpu *gpu, const struct pl_tex *tex,
+                       VkImageLayout layout, VkAccessFlags access,
+                       VkSemaphore sem_in)
+{
+    struct pl_tex_vk *tex_vk = tex->priv;
+    pl_assert(tex_vk->held);
+    pl_tex_vk_external_dep(gpu, tex, sem_in);
+
+    tex_vk->current_layout = layout;
+    tex_vk->current_access = access;
+    tex_vk->held = false;
+}
+
 // For pl_buf.priv
 struct pl_buf_vk {
     struct vk_bufslice slice;
@@ -987,6 +1075,7 @@ static const struct pl_buf *vk_buf_create(const struct pl_gpu *gpu,
         bufFlags |= VK_BUFFER_USAGE_TRANSFER_SRC_BIT |
                     VK_BUFFER_USAGE_TRANSFER_DST_BIT;
         memFlags |= VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT;
+        align = pl_lcm(align, p->min_texel_alignment);
         // Use TRANSFER-style updates for large enough buffers for efficiency
         if (params->size > 1024*1024) // 1 MB
             buf_vk->update_queue = TRANSFER;
@@ -994,28 +1083,28 @@ static const struct pl_buf *vk_buf_create(const struct pl_gpu *gpu,
     case PL_BUF_UNIFORM:
         bufFlags |= VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT;
         memFlags |= VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT;
-        align = PL_ALIGN2(align, vk->limits.minUniformBufferOffsetAlignment);
+        align = pl_lcm(align, vk->limits.minUniformBufferOffsetAlignment);
         break;
     case PL_BUF_STORAGE:
         bufFlags |= VK_BUFFER_USAGE_STORAGE_BUFFER_BIT;
         memFlags |= VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT;
-        align = PL_ALIGN2(align, vk->limits.minStorageBufferOffsetAlignment);
+        align = pl_lcm(align, vk->limits.minStorageBufferOffsetAlignment);
         buf_vk->update_queue = COMPUTE;
         break;
     case PL_BUF_TEXEL_UNIFORM: // for emulated upload
         bufFlags |= VK_BUFFER_USAGE_UNIFORM_TEXEL_BUFFER_BIT;
         bufFlags |= VK_BUFFER_USAGE_TRANSFER_DST_BIT;
         memFlags |= VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT;
-        align = PL_ALIGN2(align, vk->limits.minTexelBufferOffsetAlignment);
-        align = PL_ALIGN2(align, vk->limits.optimalBufferCopyOffsetAlignment);
+        align = pl_lcm(align, vk->limits.minTexelBufferOffsetAlignment);
+        align = pl_lcm(align, vk->limits.optimalBufferCopyOffsetAlignment);
         is_texel = true;
         break;
     case PL_BUF_TEXEL_STORAGE: // for emulated download
         bufFlags |= VK_BUFFER_USAGE_STORAGE_TEXEL_BUFFER_BIT;
         bufFlags |= VK_BUFFER_USAGE_TRANSFER_SRC_BIT;
         memFlags |= VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT;
-        align = PL_ALIGN2(align, vk->limits.minTexelBufferOffsetAlignment);
-        align = PL_ALIGN2(align, vk->limits.optimalBufferCopyOffsetAlignment);
+        align = pl_lcm(align, vk->limits.minTexelBufferOffsetAlignment);
+        align = pl_lcm(align, vk->limits.optimalBufferCopyOffsetAlignment);
         buf_vk->update_queue = COMPUTE;
         is_texel = true;
         break;
@@ -1028,7 +1117,7 @@ static const struct pl_buf *vk_buf_create(const struct pl_gpu *gpu,
 
     if (params->host_writable || params->initial_data) {
         bufFlags |= VK_BUFFER_USAGE_TRANSFER_DST_BIT;
-        align = PL_ALIGN2(align, vk->limits.optimalBufferCopyOffsetAlignment);
+        align = pl_lcm(align, vk->limits.optimalBufferCopyOffsetAlignment);
     }
 
     if (params->host_mapped || params->host_readable) {
@@ -1105,6 +1194,7 @@ static bool vk_tex_upload(const struct pl_gpu *gpu,
             .size = size,
             .format = tex_vk->texel_fmt,
         });
+
         if (!tbuf) {
             PL_ERR(gpu, "Failed creating texel buffer for emulated tex upload!");
             goto error;
@@ -1199,6 +1289,7 @@ static bool vk_tex_download(const struct pl_gpu *gpu,
             .size = size,
             .format = tex_vk->texel_fmt,
         });
+
         if (!tbuf) {
             PL_ERR(gpu, "Failed creating texel buffer for emulated tex download!");
             goto error;
@@ -1280,6 +1371,7 @@ struct pl_pass_vk {
     VkImageLayout initialLayout;
     VkImageLayout finalLayout;
     // Descriptor set (bindings)
+    bool use_pushd;
     VkDescriptorSetLayout dsLayout;
     VkDescriptorPool dsPool;
     // To keep track of which descriptor sets are and aren't available, we
@@ -1288,6 +1380,9 @@ struct pl_pass_vk {
     uint16_t dmask;
     // Vertex buffers (vertices)
     struct pl_buf_pool vbo;
+    const struct pl_buf *cached_vert;
+    void *cached_data;
+    size_t cached_size;
 
     // For updating
     VkWriteDescriptorSet *dswrite;
@@ -1420,13 +1515,17 @@ static const struct pl_pass *vk_pass_create(const struct pl_gpu *gpu,
         talloc_zero(pass, struct pl_pass_vk);
     pass_vk->dmask = -1; // all descriptors available
 
+    // temporary allocations
+    void *tmp = talloc_new(NULL);
+
     int num_desc = params->num_descriptors;
+    if (!num_desc)
+        goto no_descriptors;
+
     pass_vk->dswrite = talloc_array(pass, VkWriteDescriptorSet, num_desc);
     pass_vk->dsiinfo = talloc_array(pass, VkDescriptorImageInfo, num_desc);
     pass_vk->dsbinfo = talloc_array(pass, VkDescriptorBufferInfo, num_desc);
 
-    // temporary allocations/objects
-    void *tmp = talloc_new(NULL);
     VkPipelineCache pipeCache = NULL;
     VkShaderModule vert_shader = NULL;
     VkShaderModule frag_shader = NULL;
@@ -1450,58 +1549,69 @@ static const struct pl_pass *vk_pass_create(const struct pl_gpu *gpu,
         };
     }
 
-    VkDescriptorPoolSize *dsPoolSizes = NULL;
-    int poolSizeCount = 0;
-
-    for (enum pl_desc_type t = 0; t < PL_DESC_TYPE_COUNT; t++) {
-        if (dsSize[t] > 0) {
-            VkDescriptorPoolSize dssize = {
-                .type = dsType[t],
-                .descriptorCount = dsSize[t] * NUM_DS,
-            };
-
-            TARRAY_APPEND(tmp, dsPoolSizes, poolSizeCount, dssize);
-        }
-    }
-
-    if (poolSizeCount) {
-        VkDescriptorPoolCreateInfo pinfo = {
-            .sType = VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO,
-            .maxSets = NUM_DS,
-            .pPoolSizes = dsPoolSizes,
-            .poolSizeCount = poolSizeCount,
-        };
-
-        VK(vkCreateDescriptorPool(vk->dev, &pinfo, VK_ALLOC, &pass_vk->dsPool));
-    }
-
     VkDescriptorSetLayoutCreateInfo dinfo = {
         .sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO,
         .pBindings = bindings,
         .bindingCount = num_desc,
     };
 
+    if (p->max_push_descriptors && num_desc <= p->max_push_descriptors) {
+        dinfo.flags |= VK_DESCRIPTOR_SET_LAYOUT_CREATE_PUSH_DESCRIPTOR_BIT_KHR;
+        pass_vk->use_pushd = true;
+    } else if (p->max_push_descriptors) {
+        PL_INFO(gpu, "Pass with %d descriptors exceeds the maximum push "
+                "descriptor count (%d). Falling back to descriptor sets!",
+                num_desc, p->max_push_descriptors);
+    }
+
     VK(vkCreateDescriptorSetLayout(vk->dev, &dinfo, VK_ALLOC,
                                    &pass_vk->dsLayout));
 
-    VkDescriptorSetLayout layouts[NUM_DS];
-    for (int i = 0; i < NUM_DS; i++)
-        layouts[i] = pass_vk->dsLayout;
+    if (!pass_vk->use_pushd) {
+        VkDescriptorPoolSize *dsPoolSizes = NULL;
+        int poolSizeCount = 0;
 
-    if (pass_vk->dsPool) {
-        VkDescriptorSetAllocateInfo ainfo = {
-            .sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO,
-            .descriptorPool = pass_vk->dsPool,
-            .descriptorSetCount = NUM_DS,
-            .pSetLayouts = layouts,
-        };
+        for (enum pl_desc_type t = 0; t < PL_DESC_TYPE_COUNT; t++) {
+            if (dsSize[t] > 0) {
+                VkDescriptorPoolSize dssize = {
+                    .type = dsType[t],
+                    .descriptorCount = dsSize[t] * NUM_DS,
+                };
 
-        VK(vkAllocateDescriptorSets(vk->dev, &ainfo, pass_vk->dss));
+                TARRAY_APPEND(tmp, dsPoolSizes, poolSizeCount, dssize);
+            }
+        }
+
+        if (poolSizeCount) {
+            VkDescriptorPoolCreateInfo pinfo = {
+                .sType = VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO,
+                .maxSets = NUM_DS,
+                .pPoolSizes = dsPoolSizes,
+                .poolSizeCount = poolSizeCount,
+            };
+
+            VK(vkCreateDescriptorPool(vk->dev, &pinfo, VK_ALLOC, &pass_vk->dsPool));
+
+            VkDescriptorSetLayout layouts[NUM_DS];
+            for (int i = 0; i < NUM_DS; i++)
+                layouts[i] = pass_vk->dsLayout;
+
+            VkDescriptorSetAllocateInfo ainfo = {
+                .sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO,
+                .descriptorPool = pass_vk->dsPool,
+                .descriptorSetCount = NUM_DS,
+                .pSetLayouts = layouts,
+            };
+
+            VK(vkAllocateDescriptorSets(vk->dev, &ainfo, pass_vk->dss));
+        }
     }
 
+no_descriptors: ;
+
     VkPipelineLayoutCreateInfo linfo = {
         .sType = VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO,
-        .setLayoutCount = 1,
+        .setLayoutCount = num_desc ? 1 : 0,
         .pSetLayouts = &pass_vk->dsLayout,
         .pushConstantRangeCount = params->push_constants_size ? 1 : 0,
         .pPushConstantRanges = &(VkPushConstantRange){
@@ -1801,7 +1911,6 @@ static void vk_update_descriptor(const struct pl_gpu *gpu, struct vk_cmd *cmd,
 {
     struct pl_pass_vk *pass_vk = pass->priv;
     struct pl_desc *desc = &pass->params.descriptors[idx];
-    pl_assert(ds);
 
     VkWriteDescriptorSet *wds = &pass_vk->dswrite[idx];
     *wds = (VkWriteDescriptorSet) {
@@ -1917,6 +2026,7 @@ static void set_ds(struct pl_pass_vk *pass_vk, uintptr_t dsbit)
 static void vk_pass_run(const struct pl_gpu *gpu,
                         const struct pl_pass_run_params *params)
 {
+    struct pl_vk *p = gpu->priv;
     struct vk_ctx *vk = pl_vk_get(gpu);
     const struct pl_pass *pass = params->pass;
     struct pl_pass_vk *pass_vk = pass->priv;
@@ -1926,17 +2036,87 @@ static void vk_pass_run(const struct pl_gpu *gpu,
         [PL_PASS_COMPUTE] = COMPUTE,
     };
 
-    // Wait for a free descriptor set
-    while (!pass_vk->dmask) {
-        PL_TRACE(gpu, "No free descriptor sets! ...blocking (slow path)");
-        vk_submit(gpu);
-        vk_poll_commands(vk, 1000000); // 1ms
+    // Update the vertex buffer before dispatching the pass, do this
+    // before vk_require_cmd since it can trigger its own commands
+    const struct pl_buf *vert = NULL;
+    struct pl_buf_vk *vert_vk = NULL;
+    if (pass->params.type == PL_PASS_RASTER) {
+        size_t size = params->vertex_count * pass->params.vertex_stride;
+        if (pass_vk->cached_vert && pass_vk->cached_size == size &&
+            memcmp(params->vertex_data, pass_vk->cached_data, size) == 0)
+        {
+            // Re-use cached vertex buffer
+            vert = pass_vk->cached_vert;
+        } else {
+            // Invalidate existing cache
+            pass_vk->cached_vert = NULL;
+
+            // Fetch new vertex buffer and update it
+            vert = pl_buf_pool_get(gpu, &pass_vk->vbo, &(struct pl_buf_params) {
+                .type = PL_VK_BUF_VERTEX,
+                .size = params->vertex_count * pass->params.vertex_stride,
+                .host_writable = true,
+            });
+
+            if (!vert) {
+                PL_ERR(gpu, "Failed allocating vertex buffer!");
+                goto error;
+            }
+
+            vk_buf_write(gpu, vert, 0, params->vertex_data, vert->params.size);
+
+            // Update the cached information, for small vertex buffers
+            if (size <= 128*1024) { // 128 KiB
+                pass_vk->cached_vert = vert;
+                pass_vk->cached_size = size;
+                pass_vk->cached_data =
+                    talloc_realloc_size(pass_vk, pass_vk->cached_data, size);
+                memmove(pass_vk->cached_data, params->vertex_data, size);
+            }
+        }
+
+        pl_assert(vert);
+        vert_vk = vert->priv;
+    }
+
+    if (!pass_vk->use_pushd) {
+        // Wait for a free descriptor set
+        while (!pass_vk->dmask) {
+            PL_TRACE(gpu, "No free descriptor sets! ...blocking (slow path)");
+            vk_submit(gpu);
+            vk_poll_commands(vk, 1000000); // 1ms
+        }
     }
 
     struct vk_cmd *cmd = vk_require_cmd(gpu, types[pass->params.type]);
     if (!cmd)
         goto error;
 
+    // Find a descriptor set to use
+    VkDescriptorSet ds = NULL;
+    if (!pass_vk->use_pushd) {
+        for (int i = 0; i < PL_ARRAY_SIZE(pass_vk->dss); i++) {
+            uint16_t dsbit = 1u << i;
+            if (pass_vk->dmask & dsbit) {
+                ds = pass_vk->dss[i];
+                pass_vk->dmask &= ~dsbit; // unset
+                vk_cmd_callback(cmd, (vk_cb) set_ds, pass_vk,
+                                (void *)(uintptr_t) dsbit);
+                break;
+            }
+        }
+    }
+
+    // Update the dswrite structure with all of the new values
+    for (int i = 0; i < pass->params.num_descriptors; i++)
+        vk_update_descriptor(gpu, cmd, pass, params->desc_bindings[i], ds, i);
+
+    if (!pass_vk->use_pushd) {
+        vkUpdateDescriptorSets(vk->dev, pass->params.num_descriptors,
+                               pass_vk->dswrite, 0, NULL);
+    }
+
+    // Bind the pipeline, descriptor set, etc.
     static const VkPipelineBindPoint bindPoint[] = {
         [PL_PASS_RASTER]  = VK_PIPELINE_BIND_POINT_GRAPHICS,
         [PL_PASS_COMPUTE] = VK_PIPELINE_BIND_POINT_COMPUTE,
@@ -1944,31 +2124,18 @@ static void vk_pass_run(const struct pl_gpu *gpu,
 
     vkCmdBindPipeline(cmd->buf, bindPoint[pass->params.type], pass_vk->pipe);
 
-    VkDescriptorSet ds = NULL;
-    for (int i = 0; i < PL_ARRAY_SIZE(pass_vk->dss); i++) {
-        uint16_t dsbit = 1u << i;
-        if (pass_vk->dmask & dsbit) {
-            ds = pass_vk->dss[i];
-            pass_vk->dmask &= ~dsbit; // unset
-            vk_cmd_callback(cmd, (vk_cb) set_ds, pass_vk,
-                            (void *)(uintptr_t) dsbit);
-            break;
-        }
-    }
-
-    for (int i = 0; i < pass->params.num_descriptors; i++)
-        vk_update_descriptor(gpu, cmd, pass, params->desc_bindings[i], ds, i);
-
-    if (pass->params.num_descriptors > 0) {
-        vkUpdateDescriptorSets(vk->dev, pass->params.num_descriptors,
-                               pass_vk->dswrite, 0, NULL);
-    }
-
     if (ds) {
         vkCmdBindDescriptorSets(cmd->buf, bindPoint[pass->params.type],
                                 pass_vk->pipeLayout, 0, 1, &ds, 0, NULL);
     }
 
+    if (pass_vk->use_pushd) {
+        vk->vkCmdPushDescriptorSetKHR(cmd->buf, bindPoint[pass->params.type],
+                                      pass_vk->pipeLayout, 0,
+                                      pass->params.num_descriptors,
+                                      pass_vk->dswrite);
+    }
+
     if (pass->params.push_constants_size) {
         vkCmdPushConstants(cmd->buf, pass_vk->pipeLayout,
                            stageFlags[pass->params.type], 0,
@@ -1981,27 +2148,12 @@ static void vk_pass_run(const struct pl_gpu *gpu,
         const struct pl_tex *tex = params->target;
         struct pl_tex_vk *tex_vk = tex->priv;
 
-        struct pl_buf_params vparams = {
-            .type = PL_VK_BUF_VERTEX,
-            .size = params->vertex_count * pass->params.vertex_stride,
-            .host_writable = true,
-        };
-
-        const struct pl_buf *buf = pl_buf_pool_get(gpu, &pass_vk->vbo, &vparams);
-        if (!buf) {
-            PL_ERR(gpu, "Failed allocating vertex buffer!");
-            goto error;
-        }
-
-        struct pl_buf_vk *buf_vk = buf->priv;
-        vk_buf_write(gpu, buf, 0, params->vertex_data, vparams.size);
-
-        buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_VERTEX_INPUT_BIT,
+        buf_barrier(gpu, cmd, vert, VK_PIPELINE_STAGE_VERTEX_INPUT_BIT,
                     VK_ACCESS_VERTEX_ATTRIBUTE_READ_BIT,
-                    buf_vk->slice.mem.offset, vparams.size, false);
+                    vert_vk->slice.mem.offset, vert->params.size, false);
 
-        vkCmdBindVertexBuffers(cmd->buf, 0, 1, &buf_vk->slice.buf,
-                               &buf_vk->slice.mem.offset);
+        vkCmdBindVertexBuffers(cmd->buf, 0, 1, &vert_vk->slice.buf,
+                               &vert_vk->slice.mem.offset);
 
         tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT,
                     VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT,
@@ -2051,6 +2203,7 @@ static void vk_pass_run(const struct pl_gpu *gpu,
 
     // flush the work so far into its own command buffer, for better
     // intra-frame granularity
+    pl_assert(cmd == p->cmd); // make sure this is still the case
     vk_submit(gpu);
 
 error:
@@ -2064,20 +2217,10 @@ static void vk_gpu_flush(const struct pl_gpu *gpu)
     vk_flush_commands(vk);
 }
 
-struct vk_cmd *pl_vk_finish_frame(const struct pl_gpu *gpu,
-                                  const struct pl_tex *tex)
+struct vk_cmd *pl_vk_steal_cmd(const struct pl_gpu *gpu)
 {
     struct pl_vk *p = gpu->priv;
     struct vk_cmd *cmd = vk_require_cmd(gpu, GRAPHICS);
-    if (!cmd)
-        return NULL;
-
-    struct pl_tex_vk *tex_vk = tex->priv;
-    pl_assert(tex_vk->external_img);
-    tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
-                VK_ACCESS_MEMORY_READ_BIT, VK_IMAGE_LAYOUT_PRESENT_SRC_KHR);
-
-    // Return this directly instead of going through vk_submit
     p->cmd = NULL;
     return cmd;
 }
